;
;	RTOS7 - A Real-Time Operating System for the ARM7 CPU
;
;      (c) Copyright 2004 John B. Stephensen  All Rights Reserved
;
; This software in source, object or binary form is licensed only for
; personal non-profit educational use in the Amateur Radio Service and
; the license is not transferrable. The software is provided as-is for
; experimental purposes and the author does not warranty its freedom
; from defects or its suitability for any specific application.
;
; History:
;	v0.1	Added C language interface functions and changed from
;		integer messages to event flags. JBS 9-1-04
;	v0.2	Added SWI 8 for variable entry point to FIQ handler. 
;		JBS 10-5-04
;
; This is a simple non-preemptive real time operating system that
; executes one process at a time. When a process completes, it calls
; the OS and the OS executes the highest priority waiting process.
; One program may communicate with another by setting 32 event flags.
; Each program waits on zero or more flags. When a flag is set, the
; highest priority program will resume execution with the flags in R0.
; One program may also cede control to another program immediately,
; while passing a message in R0. The RTOS code is usually in flash ROM.
;
; This program was written for an ARM7-TDMI CPU with an ARM advanced
; interrupt controller (AIC). The CPU must be in little endian mode.
; It was tested on an Oki Semiconductor ML67Q5003 MCU.
;
; ARM7 CPU banked registers are used to track currently-running programs.
; Programs run at two levels. User-level processes may have their execution
; interrupted by interrupt service routines that are called by the OS to
; service interrupts. The fast interrupt causes execution of a short
; service routine that may interrupt user-level or system-level programs.
; The state of each process is contained in a program control block (PCB).
; Process context switching is performed when the currently-executing
; process uses a software interrupt instruction to change processor mode
; and schedule other processes for execution. Programs execute in user-mode
; and interrupt service routines execute in IRQ or FIQ mode. The operating
; system executes in SVC mode. Each mode has its own link register and
; stack pointer.
;
; Supervisor Mode - Initial state on reset. SP_spv holds address of PCB for
; currently-running user-level program. User-level programs enter this mode
; via a software interrupt instruction that requests various operating 
; system services such as posting messages, changing priorities or starting
; other programs.
;
; User Mode - Entered when background programs run. All interrupts
; (FIQ and IRQ) are enabled. Exited when interrupt serviced or
; program calls RTOS to signal another program.
;
; System Mode - Entered when interrupt service program started. FIQ is
; enabled to allow fast interrupts to be services. IRQ is disabled so
; that only one system-level program is executing at a time. Exited
; when program calls RTOS to wait for another interrupt.
;
; FIQ Mode - Entered when fast interrupt received. Executes shoert
; interrupt service routine contained in fisr.as and then returns to
; user-level or system-level program. SP_fiq points to dedicated FIQ
; stack so the the interrupted program stack is not used.
;
; IRQ Mode - Entered from user mode when interrupt active. SP_irq holds
; address of PCB for system-level program activated to service interrupt.
; Switches to user mode when system-level program waits for another
; interrupt.
;
; UND Mode - Entered when undefined instruction executed, same as abort.
; Abort Mode - Entered when memory access control violated. The offending
; program is terminated by masking FIQ (FIQ service routine), setting the
; IRL to zero (system-level program) or blocking the message queue (user-
; level program). The code then attempts to return to the previously-
; executing program. If this was the RTOS, there is a fatal bug and the
; CPU is reset.
;
; Usage of six stack pointer registers:
;
;	FIQ - holds FIQ ISR address
;	IRQ - holds PCB address for current system-mode program
;	SVC - holds PCB address for current user-mode program
;	ABT - top of stack for abort routine
;	UND - top of stack for abort routine
;	USR - top of stack for user-mode and system-mode programs
;
; The FIQ ISR address is in SP_fiq. If it is zero, the interrupt is
; not serviced. The service routine must use only registers R8-R12
; and CPSR_fiq. The ISR may change entry points by loading a new
; value into SP_fiq. The following code causes the next interrupt
; to return to the instruction following the return from interrupt.
;  
;	; ISR code segment 1
;	...
;	...
;	MOV	SP, PC	; PC+8 -> SP switches to code segment 2
;	MOVS	PC, LR	; return to interrupted program
;	; ISR code segment 2
;	...
;
; The FIQ ISR has no stack and cannot call another routine. It must be
; short and fast. Otherwise, a regular interrupt should be used.
;
; There are three tables that are used for program scheduling:
;
;	ISRTBL - 32 entries, indexed by IRN, pointing to ISR programs
;	PCBTBL - 16 entries, indexed by priority, pointing to user program PCBs
;	EVTTBL - 16 entries, indexed by priority, containing event flag masks
;
; When there is no activity, the idle loop checks the event flag word to see
; if any bits are set. If any are set, it scans the PCBTBL for user programs
; that are monitoring the set flags. This is determined by ANDing the flags
; and the event flag mask in EVTTBL. The highest priority program is started
; by loading its registers from its PCB. When the program is done, it calls
; the OS and its registers are saved to the PCB. Interrupt service routines
; execute when an interrupt occurs. The user program state is saved on the
; interrupt stack (IRQSTK) and the ISR registered for that interrupt number
; is executed by loading its registers from its PCB. When the system program
; is done it returns to the OS and the interrupted user program state is
; restored from IRQSTK and it continues execution.
;
; There are 8 OS calls that use the software interrupt instruction:
;
;	SWI 0 - Resume execution of interrupted or lower-priority program
;	SWI 1 - Register user program & set priority and event flag mask
;	SWI 2 - Change user program event flag mask
;	SWI 3 - Exit user program and start another specified user program
;	SWI 4 - Register system program & set interrupt number and priority
;	SWI 5 - Change system program interrupt priority
;	SWI 6 - Set an event flag
;	SWI 7 - Register fast interrupt service routine
;
; The OS uses registers R0-R3 for parameters and R12 is modified. All other
; registers are preserved. R0 is used to return values. When programs are
; resumed, it contains IRN (interrupt number) for system programs and the
; monitored event flags that were set for user programs. R0 is zero if a
; user program execution resumes because of a SWI 3 OS call. 
;
; The process control blocks (PCBs) for the user level programs are
; located at the top of the space allocated to the programs for
; stack. It contains the program state informetion as follows:
;
;	PCB-00:	R15/PC - Program Counter of interrupted program
;	PCB-04:	R14/LR - Link Register containing return address
;	PCB-08:	R13/SP - Stack Pointer that points to top of stack
;	PCB-0C:	R12 - all other general registers
;	PCB-10:	R11
;	PCB-14:	R10
;	PCB-18:	R9
;	PCB-1C:	R8
;	PCB-20:	R7
;	PCB-24:	R6
;	PCB-28:	R5
;	PCB-2C:	R4
;	PCB-30:	BOS - Bottom of stack that grows down in memory
USRPC	EQU	0x00
USRLR	EQU	0x04
USRSP	EQU	0x08
USRBOS	EQU	0x30
;
; The following point to external C-language functions and variables
;
	GLOBAL	_osabort	; address of abort notification function
	GLOBAL	start		; starting address of main program
	GLOBAL	_osclock	; system clock word address;
;
; The following are for C language OS calls
;
	GLOBAL	_oswait		; wait for message
	GLOBAL	_osregpgm	; register user program
	GLOBAL	_ossetmsk	; set event flag mask
	GLOBAL	_osresume	; transfer control to another user program
	GLOBAL	_osregisr	; register system program (ISR)
	GLOBAL	_oschgirl	; change interrupt priority (level)
	GLOBAL	_ossetflg	; set a flag from user program
	GLOBAL	_osregfiq	; register FIQ ISR
;
; Data segment used by this program to go in bottom of ML67Q500x RAM
;
        PSECT	bss, class=RAM
;	ORG	0x50000000
;
EVTFLG:	DS	4	; 32 event flag bits
ISRTBL:	DS	128	; table of pointers to IRQ service routines by IRN
PCBTBL:	DS	64	; table of pointers to PCBs by priority
EVTTBL:	DS	64	; table of event flag masks by PCB priority
	DS	64	; FIQ stack - 16 entries max.
FIQSTK:	DS	4	; BOS for FIQ service routine
	DS	64	; PSR & R0-R14 for interrupted programs
IRQSTK:	DS	4	; saved PC for interrupted programs
	DS	64	; abort stack - 16 entries max.
ABTSTK:	DS	4	; BOS for abort service routine 
;
; Constants used in this program
;
PCBMAX	EQU	16	; maximum number of PCBs
USR	EQU	0x10	; user mode
FIQ	EQU	0x11	; FIQ mode
IRQ	EQU	0x12	; IRQ mode
SVC	EQU	0x13	; supervisor call mode
ABT	EQU	0x17	; abort mode
UND	EQU	0x1B	; undefined instruction mode
SYS	EQU	0x1F	; system mode
FIQM	EQU	0x40	; mask FIQ interrupts
IRQM	EQU	0x80	; mask IRQ interrupts
;
; start code at reset vector location
;
        PSECT	vectors, class=CODE
	ORG	0
;
; Exception Vectors - branch to code to handle each type
;
	B	RESET	; MCU reset
	B	UNDEF	; undefined instruction exception*
	B	SWINT	; software interrupt request
	B	PABORT	; prefect abort exception
	B	DABORT	; data abort exception
	B	IDLE	; reserved vector - go to idle loop
	B	IRQSVC	; hardware interrupt request
;
; Fast Interrupt Service Routine (FIQ ISR)
;
FIQSVC:
	SUB	LR, LR, #4	; fix return address
	CMP	SP, #0		; check if ISR registered
	MOVNE	PC, SP		; set PC if so
	MOVS	PC, LR		; else return from interrupt
;
; Hardware Interrupt Service Routine
;
; Save state of current program on IRQ stack, get interrupt nummber and
; resume program associated with highest-priority active interrupt.
;	SP_irq is pointer to PCB of program to service interrupt
;
IRQSVC:
	;save registers for interrupted program
        LOAD    SP, #IRQSTK	; BOS for ISR
	SUB	LR, LR, #4	; adjust return address
	STMFD   SP!, {R0-R12,LR}	; save registers
        ; Did we get an interrupt while interrupts were disabled? If so, 
        ; return immediately. The interrupt will remain pending.
	MRS	R12, SPSR	; get saved status register
        TST     R12, #0x80      ; IRQ inhibit bit in SPSR bit set ?
        BNE     IRQRTN          ; If so, just return immediately.
        ; Get IRQ number from the IRN register, and save it into R0
        LOAD    R12, #0x78000000 ; R12 = Interrupt controller base address
        LDR     R0, [R12, #0x14] ; R0 = IRQ number from the IRN register
        ; Check IRQ number - if invalid (irn > IRQSIZE) ignore the interrupt
        CMP     R0, #0x20	; is IRN > IRQSIZE ?
        BCS     IRQRTN          ; if invalid, just return immediately.
	; check for system timer interrupt 
	CMP	R0, #0		; is IRN 0 ?
	BEQ	SYSTIM		; jump if yes
        ; retreive address of ISR TCB from isr_tbl + irn*4 and call it
        LOAD    R12, #ISRTBL	; R12 = Table of PCB addresses
        LDR     R12, [R12, R0, LSL #2] ; get ISR address indexed by IRN
	; turn on red LED
	LOAD	R4, #0xB7A01060	; GPPOD
	MOV	R5, #0x08	; red
	STRH	R5, [R4]	; write
	; call ISR
	; check for NUL
	CMP	R12, #0		; NUL = 0
	MOV	LR, PC		; PC+8 -> LR
	BXNE	R12		; ISR address -> PC
	; turn on green LED
	LOAD	R4, #0xB7A01060	; GPPOD
	MOV	R5, #0x04	; green
	STRH	R5, [R4]	; write
IRQDONE:
	; write to CICL to clear highest 1 bit in CIL
        LOAD    R12, #0x78000000 ; R12 = Interrupt controller base address
        STR     R12, [R12, #0x28] ; write to CICL (data value ignored)
IRQRTN:
	; restore registers with LR -> PC and SPSR -> CPSR
	LDMFD	SP!, {R0-R12,PC}^	; return from interrupt
;
; System timer interrupt service routine
;
SYSTIM:
	; here if system timer interrupt or AIC error
	LOAD	R12, #0xB8001000 ; system timer base
	LDR	R11, [R12, #0x10] ; load overflow register
	ANDS	R11, R11, #0x01	; check overflow bit
	BEQ	IRQRTN		; return if not set (error)
	STR	R11, [R12, #0x10] ; reset timer interrupt
	LOAD	R12, #_osclock	; increment clock
	LDR	R11, [R12]
	ADD	R11, R11, #1
	STR	R11, [R12]
	B	IRQDONE		; done
;
; Software Interrupt Service Routine
;	SP_svc is pointer to calling program's PCB
;	SWI 0 is wait for event command
;	Other commands decoded in SWISVC
;	Modifies R0-R3
; Uses (after saving context):
;	R0 = event flag values
;	R4 = event flags address
;	R5 = mask/PCB table base address
;	R6 = table index (priority)
;	R7 = event flag mask
;	R12 = temp
; Returns (to resumed program):
;	R0 = flags monitored by program
SWINT:
	; decode SWI comment field
	LDR	R12, [LR, #-4]	; load SWI instruction
	ANDS	R12, R12, #0x0F	; mask all but lower 4 bits
	BNE	SWISVC		; if not zero, decode command
	; save state of calling program - SP points to current PCB
	STR     LR, [SP]        ; save return address
	STMDB   SP, {R4-R14}^	; save user-mode registers
IDLE:
	; turn on green LED
	LOAD	R4, #0xB7A01060	; GPPOD
	MOV	R5, #0x04	; green
	STRH	R5, [R4]	; write
IDLELOOP:
	; enable interrupts
	MRS	R12, CPSR	; get status
	BIC	R12, R12, #FIQM|IRQM	; enable IRQ and FIQ
	MSR	CPSR_c, R12	; set status
	; check event flags
	LOAD	R4, #EVTFLG	; get address of flag word
CHKFLGS:
	LDR	R0, [R4]	; get flags
	CMP	R0, #0		; set status register
	BEQ	CHKFLGS		; loop if none set
	; disable interrupts
	MRS	R12, CPSR	; get status
	ORR	R12, R12, #FIQM|IRQM	; disable IRQ and FIQ
	MSR	CPSR_c, R12	; set status
	; determine program to run next by checking flags
	LDR	R0, [R4]	; get flags (in case of interrupt)
	LOAD	R5, #EVTTBL	; event mask table base
	MOV	R6, #PCBMAX	; number of entries
IDLESCAN:
	; check for unmasked flags
	LDR	R7, [R5, R6, LSL #2]	; get mask
	ANDS	R12, R0, R7	; apply mask
	BNE	EVENT		; jump if any set
	SUBS	R6, R6, #1	; decrement index
	BNE	IDLESCAN	; check next program
	B	IDLELOOP	; no waiting program, try again
EVENT:
	; flag(s) set - run program indexed by R5
	BIC	R7, R0, R7	; clear masked flags
	STR	R7, [R4]	; and save result
	MOV	R0, R12		; make monitored flags available
	LOAD	R5, #PCBTBL	; PCB address table
	LDR	SP, [R5, R6, LSL #2] ; load PCB address
	; check that program exists
	CMP	SP, #0		; if no program
	BEQ	IDLELOOP	; wait for another event
RUNUSR:
	; turn on red LED
	LOAD	R4, #0xB7A01060	; GPPOD
	MOV	R5, #0x08	; red
	STRH	R5, [R4]	; write
	; run user-mode program - SP points to current PCB
	MSR     SPSR_c, #USR  	; set user-mode PSR, all int.
        LDMDB   SP, {R4-R14}^   ; restore user-mode registers
	NOP			; CPU bug fix
	LDMDA   SP, {PC}^       ; LR -> PC, SPSR -> CPSR and go
;
; SWI error: return -1
;
SWIERR:
	LOAD	R0, #-1		; return error code
SWIRTN:
	MOVS	PC, LR		; return with CPSR <- SPSR
;
; Prefetch abort, data abort and undefined instruction handler
;
DABORT:
	SUB	LR, LR, #4	; adjust PC (-8 total)
PABORT:
UNDEF:
	SUB	LR, LR, #4	; adjust PC
	LOAD	SP, #ABTSTK	; stack for C-language abort handler
	STMFD	SP!, {LR}	; save user registers
	STMFD	SP, {R0-R14}^
	NOP
	SUB	SP, SP, #60	; adjust stack pointer
	MOV	R0, SP		; pointer for abort routine
	MRS	R8, SPSR	; get saved processor mode	
	AND	R8, R8, #0x1F	; mask out all but mode bits
	CMP	R8, #USR	; user mode
	BEQ	SYSERR		; jump to handler
	CMP	R8, #FIQ	; FIQ mode
	BEQ	FIQERR		; jump to handler
	CMP	R8, #IRQ	; IRQ mode
	BEQ	IRQERR		; jump to handler
	CMP	R8, #SYS	; system mode
	BEQ	SYSERR		; jump to handler
;
; aborted in supervisor mode
; or two aborts in a row
;
SPVERR:
	MOV	R1, LR		; get CPU state
	MRS	R2, SPSR
	MRS	R3, CPSR
	BL	_osabort	; call C error-handling routine
	B	RESET		; no recovery - RESET
;
; Here if abort occurred during fast interrupt service routine
;
FIQERR:	
	; stop fast interrupts
        LOAD    R8, #0x78000000 ; R8 = Interrupt controller base address
	MOV	R9, #0		; reset FIQ enable bit
        STR     R9, [R8, #0x10]	; write to FIQ enable register
	; call C-code error handler
	MOV	R1, LR		; get CPU state
	MRS	R2, SPSR
	MRS	R3, CPSR
	BL	_osabort
	LDMFD	SP!,{R0-R7}	; restore user registers
	; change to FIQ mode
	MRS	R8, SPSR	; get saved PSR
	MSR	CPSR_cxsf, R8	; make current PSR
	; return to interrupted code
	SUBS	PC, LR, #4	; return to interrupted program	
;
; Here if aborted during interrupt service routine execution
;
IRQERR:
	; report error
	MOV	R1, LR		; get CPU state
	MRS	R2, SPSR
	MRS	R3, CPSR
	BL	_osabort	; call error-handling routine
	; change to IRQ mode with no interrupts
	MSR	CPSR_c, #IRQ|FIQM|IRQM
	; done - restore interrupted program
        LOAD    SP, #IRQSTK-56	; TOS when ISR called
	B	IRQDONE		; return from interrupt
;
; Here if aborted during program execution
;
SYSERR:
USRERR:
	; report error
	MOV	R1, LR		; get CPU state
	MRS	R2, SPSR
	MRS	R3, CPSR
	BL	_osabort	; call error-handling routine
	; change to supervisor mode with no interrupts
	MSR	CPSR_c, #SVC|FIQM|IRQM
	; disable aborted program 
	LOAD	R4, #PCBTBL	; PCB table
	LOAD	R6, #EVTTBL	; event mask table
	LOAD	R5, #0		; reset index
	LOAD	R7, #0		; mask all flags value
USRPCB:
	LDR	R8, [R4, R5, LSL #2] ; get PCB address
	CMP	R8, SP		; compare aborted PCB address
	STREQ	R7, [R4, R5, LSL #2] ; disable if aborted PCB
	STREQ	R7, [R6, R5, LSL #2] ; disable if aborted PCB
	ADD	R5, R5, #1	; increment index
	CMP	R5, #PCBMAX	; check for end of table
	BLT	USRPCB		; if not, check another entry
	; done - go to idle loop
	B	IDLE		; run next program
;
; Decode SWI command in R12
;
MAXCMD	EQU	8
SWISVC:
	CMP	R12, #MAXCMD	; check if valid command
	LDRLO	PC, [PC, R12, LSL #2] ; jump through table if valid
	B	SWIERR		; return if not valid
	DDW	SPVERR		; should never be accessed
	DDW	REGPGM		; SWI 1 - register user program
	DDW	SETMSK		; SWI 2 - set event flag mask
	DDW	RESUME		; SWI 3 - resume selected program
	DDW	REGISR		; SWI 4 - register system program (ISR)
	DDW	CHGIRL		; SWI 5 - change interrupt level
	DDW	SETFLG		; SWI 6 - set one event flag
	DDW	REGFIQ		; SWI 7 - register FIQ service routine
;
; Register user program in PCB table
;	R0 = priority
;	R1 = event flag mask (0 if program to idle)
;	R2 = PCB address
;	R3 = PC start address
; Uses R12.
;
REGPGM:
	; check parameters
	CMP	R0, #PCBMAX	; check no. PCBs
	BGE	SWIERR		; return if error
	; initialize PCB
	STR	R3, [R2, #-USRPC] ; save PC
	STR	R3, [R2, #-USRLR] ; save PC in LR
	SUB	R3, R2, #USRBOS	; calculate BOS
	STR	R3, [R2, #-USRSP] ; set SP at BOS
	; make entry in PCB table
	LOAD	R12, #PCBTBL	; PCB address table
	STR	R2, [R12, R0, LSL #2] ; save PCB address
;
; Change event flags
;	R0 = priority
;	R1 = event flag mask (0 if program to idle)
; Uses R12.
;
SETMSK:
	; check parameters
	CMP	R0, #PCBMAX	; check no. PCBs
	BGE	SWIERR		; return if error
	; set entry in event mask table
	LOAD	R12, #EVTTBL	; message table
	STR	R1, [R12, R0, LSL #2] ; save mask
	MOV	R0, #0		; success code
	MOVS	PC, LR		; and return
;
; Resume program
;	R0 = target program priority or PCB address
; 	Uses R12.
;
RESUME:
	; save state of calling program - SP points to current PCB
	STR     LR, [SP]        ; save return address
	STMDB   SP, {R4-R14}^	; save user-mode registers
	; run program specified by PCB address in R1
	MOV	SP, R0		; SP = PCB address or priority
	MOV	R0, #0		; signal program no event flag set
	CMP	SP, #PCBMAX	; check which
	BGE	RUNUSR		; jump if PCB address
	; run program specified by priority in R1
	LOAD	R12, #PCBTBL	; PCB address table
	LDR	SP, [R12, SP, LSL #2] ; get PCB address
	CMP	SP, #0		; check if NUL
	BNE	RUNUSR		; run program if not NUL
	; no program to run - wait for event
	B	IDLE
;
; Register ISR program in ISR table
;	R0 = IRN (0-31)
;	R1 = IRL (0 to disable interrupt)
;	R2 = ISR address
; Uses R12
;
REGISR:
	; check IRN
	CMP	R0, #31		; IRN = 0-31
	BGT	SWIERR		; return if out of range
	CMP	R1, #7		; IRL = 0-7
	BGT	SWIERR		; return if out of range
	; make entry in PCB table
	LOAD	R12, #ISRTBL	; ISR PCB address table
	STR	R2, [R12, R0, LSL #2] ; save PCB address
;
; Change interrupt level (priority)
;   R0 = IRN
;   R1 = IRL (not modified)
; Uses:
;   R2 = register base address
;   R3 = register offset
;   R12 = rotate amount
;   R0 = ILC register value
;
CHGIRL:
	; check IRN and IRL
	CMP	R0, #31		; IRN = 0-31
	BGT	SWIERR		; return if out of range
	CMP	R1, #7		; IRL = 0-7
	BGT	SWIERR		; return if out of range
	; calculate ILC register address
	LOAD	R2, #0x7BF00000	; use ILC if IRN = 16-31
	MOV	R3, #0x18
	CMP	R0, #16		; use ILC1 if IRN = 8-15
	LOADLT	R2, #0x78000000
	LOADLT	R3, #0x24
	CMP	R0, #8		; use ILC0 if IRN = 0-7
	LOADLT	R3, #0x20
	; calculate ROR amount for IRL bits for IRN
	MOV	R12, R0
	CMP	R0, #16		; IRNs paired above IRN15
	MOVGE	R12, R12, ASR #1	; so halve rotate amount
	MOV	R12, R12, LSL #2	; rotate 4-bits ata time
	; modify register value
	LDR	R0, [R2, R3]	; load ILC register contents
	MOV	R0, R0, ROR R12 ; rotate desired field into place
	BIC	R0, R0, #0x0F	; clear lower 4 bits
	ORR	R0, R0, R1	; or in IRL
	RSB	R12, R12, #32	; shift bits back into place
	MOV	R0, R0, ROR R12
	STR	R0, [R2, R3]	; store into ILC
	MOV	R0, #0		; success code
	MOVS	PC, LR		; return from SWI
;
; Set one event flag
;	R0 = flag number to set
; Uses R1, R2 and R12.
;
SETFLG:
	; check parameters
	AND	R0, R0, #0x1F	; 0-31 allowed
	LOAD	R12, #EVTFLG	; flag address
	LDR	R1, [R12]	; get flags
	MOV	R1, R1, ROR R0	; shift flag into bit 0
	ORR	R1, R1, #1	; set flag
	RSB	R0, R0, #32	; 32-bit total shift
	MOV	R1, R1, ROR R0	; shift back
	STR	R1, [R12]	; save flags
	MOVS	PC, LR		; and return
;
; Register FIQ service routine address
;	R0 = address
; Uses R12
;
REGFIQ:
	MRS	R12, CPSR	; save current state
	MSR	CPSR_c, #FIQ|IRQM|FIQM	; go to FIQ state
	MOV	SP, R0		; set SP_fiq
	MSR	CPSR_c, R12	; restore state
	MOVS	PC, LR		; return
;
; MCU RESET - CPU in supervisor mode
;
RESET:
	; setup ROM access timing for 30-60 MHz clock
	LOAD	R0, #3		; AS=1, OE/WE=5, RO=3, BT=3
	LOAD	R1, #0x78100004	; ROMAC
	STR	R0, [R1]	; set ROM access timing
	; reset FIQ ISR pointer
	MSR	CPSR_c, #FIQ|IRQM|FIQM	; go to FIQ state
	MOV	SP, #0		; reset SP_fiq
	; make sure CPU in supervisor mode
	MSR	CPSR_c, #SVC|FIQM|IRQM
	; reset event flags
	LOAD	R0, #0
	LOAD	R1, #EVTFLG
	STR	R0, [R1]
	; make sure internal flash mapped to bottom of memory
;	LOAD	R0, #0
;	LOAD	R1, #0xB8000000
;	STR	R0, [R1, #0x10]
;	LOAD	R1, #0xB7000000
;	STR	R0, [R1, #0x0C]
	; configure chip power control for ML67x00x series
;	LOAD	R1, #0xB7000000	; BCKCTL base
;	LOAD	R0, #0x1102	; disable power to PWM, DRAM, I2C
;	STR	R0, [R1, #0x04]	; store in BCKCTL register
	; configure clock speeds
;	LOAD	R1, #0xB7000000	; CLKSTP base
;	LOAD	R0, #0x3C	; write enable
;	STR	R0, [R1, #0x04]	; CLKSTP register
;	LOAD	R0, #0x00000000	; supply clocks / normal operation
;	STR	R0, [R1, #0x04]	; store in CLKSTP register
;	LOAD	R0, #0x3C	; write enable
;	STR	R0, [R1, #0x08]	; CGBCMT0 register
;	LOAD	R0, #0x00	; HCLK div 1 & CCLK div 1
;	STR	R0, [R1, #0x08]	; store in CGBCMT0 register
;	LOAD	R0, #0x3C	; write enable
;	STR	R0, [R1, #0x0C]	; CKWT register
;	LOAD	R0, #0x0F	; 10-24 ms wait on power up
;	STR	R0, [R1, #0x0C]	; store in CKWT register
	; mask all interrupts in AIC
	MOV	R0, #0		; reset all IRLs to 0
        LOAD    R1, #0x78000000 ; Interrupt controller base address
        STR     R0, [R1, #0x10]	; write to FIQ enable register
	STR	R0, [R1, #0x20]	; ILC0: IRN0-7
	STR	R0, [R1, #0x24]	; ILC1: IRN8-15
	LOAD	R1, #0x7BF00000	; 2nd half
	STR	R0, [R1, #0x18]	; ILC: IRN16-31
	; configure system timer
	LOAD	R1, #0xB8001000 ; system timer base
	LOAD	R0, #-3456	; 1 ms with 55.296 MHz clock
	STR	R0, [R1, #0x08]	; set reload register
	LOAD	R0, #0x01	; enable bit
	STR	R0, [R1, #0x04]	; set timer enable register
	LOAD	R1, #0x78000000	; AIC base
	MOV	R0, #0x07	; ILR0 = 7
	STR	R0, [R1, #0x20]	; set ILC0
	; configure LED port
	LOAD	R1, #0xB7A01000	; PIO port D base
	MOV	R0, #0x000C	; bits 2 & 3 are output
	STRH	R0, [R1, #0x68]	; GPPMD
	MOV	R0, #0x0000	; no interrupts
	STRH	R0, [R1, #0x6C]	; GPPIED
	MOV	R0, #0x04	; LED green
	STRH	R0, [R1, #0x60]	; GPPOD
	; initialize and run main program
	B	start
;
; Entry points for C-language function calls
;
_oswait:
	SWI	0
	MOV	PC, LR
_osregpgm:
	SWI	1
	MOV	PC, LR
_ossetmsk:
	SWI	2
	MOV	PC, LR
_osresume:
	SWI	3
	MOV	PC, LR
_osregisr:
	SWI	4
	MOV	PC, LR
_oschgirl:
	SWI	5
	MOV	PC, LR
_ossetflg:
	SWI	6
	MOV	PC, LR
_osregfiq:
	SWI	7
	MOV	PC, LR
;
        END

