;
;             CIRCULARBUFFERS.AS - Circular buffer handling
;
;      (c) Copyright 2004 John Stephensen  All Rights Reserved
;
; This software in source, object or binary form is licensed only for
; personal non-profit educational use in the Amateur Radio Service and
; the license is not transferrable. The software is provided as-is for
; experimental purposes and the author does not warranty its freedom
; from defects or its suitability for any specific application.
;
; This modules consists of a set of assembly language routines,
; callable from C, that manipulte circular buffers. The buffer
; consists of four buffer control objects and a number of bytes
; available for storage. All routines are protected from interrupts
; by disabling FIQ and IRQ on entry and restoring the previous 
; interrupt masks befor exit. The circular buffer structure is 
; shown below.
;
;	+----------------+
;	|      SIZE      | 0
;	+----------------+
;	|      FIRST     | 4
;	+----------------+
;	|      STOP      | 8
;	+----------------+
;	|      NEXT      | 12
;	+----------------+
;	|                | 16
;	|      DATA      |
;	|                |
;	+----------------+
;
; The size field is initialized to the number of bytes in the buffer,
; including the four buffer control objects. It may not be modified
; during use. The first field is the offset to the first byte in the
; buffer. The stop field is the offset to the location following the
; last readable byte in the buffer. The next field is he offset to
; the next location that will be written. There are 12 functions
; that operate on circular buffers. All are called with the address
; of the buffer in R0:
;
;	cbinit - initialize the buffer
;	cbchkb - returns number of readable 8-bit bytes in buffer
;	cbchkh - returns number of readable 16-bit half-words in buffer
;	cbchkw - returns number of readable 32-bit words in buffer
;	cbgetb - returns one 8-bit byte from buffer
;	cbputb - places one 8-bit byte into buffer
;	cbgeth - returns one 16-bit half-word from buffer
;	cbputh - places one 16-bit half-word into buffer
;	cbgetw - returns one 32-bit word from buffer
;	cbputw - places one 32-bit word into buffer
;	cbcommit - makes data written since last commit readable
;	cbrollback - removes data written since last commit
;
; Use either 8, 16 or 32-bit access exclusively on read or write to
; avoid word misalignment problems. All buffers contain an integer
; number of 32-bit words, even if they must be up to 3 bytes shorter
; than specified in cbinit(). cbputx() returns non-zero if the write
; was successfull. cbgetx() returns an error code in R1 so, if C-
; language calls are used, cbchkx() must be called initially to
; determine how many data points can be read. x is b for byte (8-bit)
; bit) access, h for half-word (16-bit) access and w for word (32-bit)
; access. Circular buffers must be aligned with word boundaries.
;
	PSECT	text, class=CODE, reloc=4
	GLOBAL	_cbinit
	GLOBAL	_cbchkb
	GLOBAL	_cbchkh
	GLOBAL	_cbchkw
	GLOBAL	_cbgetb
	GLOBAL	_cbputb
	GLOBAL	_cbgeth
	GLOBAL	_cbputh
	GLOBAL	_cbgetw
	GLOBAL	_cbputw
	GLOBAL	_cbcommit
	GLOBAL	_cbrollback
;
; cbinit(char *buffer, int size)
;
; Called with:
;	R0 = buffer address
;	R1 = buffer size in words (including control objects)
; Uses:
;	R1 = temp
;	R3 = temp
;	R12 = status flags
;
_cbinit:
	CMP	R0, #0		; check for null address
	CMPNE	R1, #19		; check for adequate size
	MOVLE	PC, LR		; return if error
	MOV	R1, R1, LSL #2	; words -> bytes
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	STR	R1, [R0]	; save buffer size
	MOV	R1, #16		; data field offset
	STR	R1, [R0, #4]	; set first
	STR	R1, [R0, #8]	; set stop
	STR	R1, [R0, #12]	; set next
	MSR	CPSR_c, R12	; restore status flags
	MOV	PC, LR		; return
;
; int cbchkb(char *buffer)
; int cbchkh(char *buffer)
; int cbchkw(char *buffer)
;
; Called with:
;	R0 = buffer address
; Uses:
;	R1 = data length
;	R2 = first byte index
;	R3 = stop index & buffer size
;	R12 = status flags
; Returns:
;	R0 = data available to read
;
_cbchkw:
	MOV	R1, #2		; return words
	B	cbchk
_cbchkh:
	MOV	R1, #1		; return half-words
	B	cbchk
_cbchkb:
	MOV	R1, #0		; return bytes
cbchk:
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	LDR	R2, [R0, #4]	; first
	LDR	R3, [R0, #8]	; stop
	SUBS	R0, R3, R2	; stop - first
	LDRLT	R3, [R0]	; size
	ADDLT	R0, R0, R3	; add size if negative
	SUBLTS	R0, R0, #16	; adjust for offset
	MSR	CPSR_c, R12	; restore status flags
	MOV	R0, R0, LSR R1	; adjust for data length
	MOV	PC, LR		; return
;
; char cbgetb(char *buffer)
; short cbgeth(char *buffer)
; int cbgetw(char *buffer)
;
; Called with:
;	R0 = buffer address
; Uses:
;	R2 = first byte index
;	R3 = stop index / buffer size
;	R12 = status flags
; Returns:
;	R0 = retreived data
;	R1 = non-zero if success
;
_cbgetb:
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	LDR	R2, [R0, #4]	; first
	LDR	R3, [R0, #8]	; stop
	SUBS	R1, R2, R3	; any space?
	BEQ	cbgetb0		; no, return
	LDRB	R3, [R0, R2]	; get char
	ADD	R2, R2, #1	; increment index
	LDR	R3, [R0]	; size
	CMP	R2, R3		; first = size?
	MOVEQ	R2, #16		; if so, reset index
	STR	R2, [R0, #4]	; save first
	MOV	R0, R3		; set result
cbgetb0:
	MSR	CPSR_c, R12	; restore status flags
	MOV	PC, LR		; return
_cbgeth:
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	LDR	R2, [R0, #4]	; first
	LDR	R3, [R0, #8]	; stop
	SUBS	R1, R2, R3	; any space?
	BEQ	cbgeth0		; no, return
	LDRH	R3, [R0, R2]	; get half-word
	ADD	R2, R2, #2	; increment index
	LDR	R3, [R0]	; size
	CMP	R2, R3		; first = size?
	MOVEQ	R2, #16		; if so, reset index
	STR	R2, [R0, #4]	; save first
	MOV	R0, R3		; set result
cbgeth0:
	MSR	CPSR_c, R12	; restore status flags
	MOV	PC, LR		; return
_cbgetw:
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	LDR	R2, [R0, #4]	; first
	LDR	R3, [R0, #8]	; stop
	SUBS	R1, R2, R3	; any space?
	BEQ	cbgetw0		; no, return
	LDR	R3, [R0, R2]	; get word
	ADD	R2, R2, #4	; increment index
	LDR	R3, [R0]	; size
	CMP	R2, R3		; first = size?
	MOVEQ	R2, #16		; if so, reset index
	STR	R2, [R0, #4]	; save first
	MOV	R0, R3		; set result
cbgetw0:
	MSR	CPSR_c, R12	; restore status flags
	MOV	PC, LR		; return
;
; int cbputb(char *buffer, char data)
; int cbputh(char *buffer, short data)
; int cbputw(char *buffer, int data)
;
; Called with:
;	R0 = buffer address
;	R1 = data to place in buffer
; Uses:
;	R2 = buffer size / first byte index
;	R3 = next byte index
;	R12 = status flags
; Returns:
;	R0 = non-zero if success
;
_cbputb:
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	LDR	R2, [R0]	; size
	LDR	R3, [R0, #12]	; next
	ADD	R3, R3, #1	; increment index
	CMP	R3, R2		; next = size?
	MOVEQ	R3, #16		; if so, reset index
	LDR	R2, [R0, #4]	; first
	SUBS	R2, R3, R2	; next = first?
	STRNE	R3, [R0, #12]	; if not, save next
	STRNEB	R1, [R0, R3]	; if not, save data
	MSR	CPSR_c, R12	; restore status flags
	MOV	R0, R2		; non-zero if success
	MOV	PC, LR		; return
_cbputh:
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	LDR	R2, [R0]	; size
	LDR	R3, [R0, #12]	; next
	ADD	R3, R3, #2	; increment index
	CMP	R3, R2		; next = size?
	MOVEQ	R3, #16		; if so, reset index
	LDR	R2, [R0, #4]	; first
	SUBS	R2, R3, R2	; next = first?
	STRNE	R3, [R0, #12]	; if not, save next
	STRNEH	R1, [R0, R3]	; if not, save data
	MSR	CPSR_c, R12	; restore status flags
	MOV	R0, R2		; non-zero if success
	MOV	PC, LR		; return
_cbputw:
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	LDR	R2, [R0]	; size
	LDR	R3, [R0, #12]	; next
	ADD	R3, R3, #4	; increment index
	CMP	R3, R2		; next = size?
	MOVEQ	R3, #16		; if so, reset index
	LDR	R2, [R0, #4]	; first
	SUBS	R2, R3, R2	; next = first?
	STRNE	R3, [R0, #12]	; if not, save next
	STRNE	R1, [R0, R3]	; if not, save data
	MSR	CPSR_c, R12	; restore status flags
	MOV	R0, R2		; non-zero if success
	MOV	PC, LR		; return
;
; cbcommit(char *buffer)
;
; Called with:
;	R0 = buffer address
; Uses:
;	R3 = temp
;	R12 = status flags
;
_cbcommit:
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	LDR	R3, [R0, #12]	; next
	STR	R3, [R0, #8]	; stop
	MSR	CPSR_c, R12	; restore status flags
	MOV	PC, LR		; return
;
; cbrollback(char *buffer)
;
; Called with:
;	R0 = buffer address
; Uses:
;	R3 = temp
;	R12 = status flags
;
_cbrollback:
	MRS	R12, CPSR	; save status flags
	ORR	R3, R12, #0xC0	; set interrupt masks
	MSR	CPSR_c, R3	; set status flags
	LDR	R3, [R0, #8]	; stop
	STR	R3, [R0, #12]	; next
	MSR	CPSR_c, R12	; restore status flags
	MOV	PC, LR		; return
;

