/*
 *		FPGA Configuration Download Program
 *
 *      (c) Copyright 2004 John B. Stephensen  All Rights Reserved
 *
 * This software in source, object or binary form is licensed only for
 * personal non-profit educational use in the Amateur Radio Service and
 * the license is not transferrable. The software is provided as-is for
 * experimental purposes and the author does not warranty its freedom
 * from defects or its suitability for any specific application.
 *
 * This file contains a program to remotely control the JTAG port. 
 * jtag.h refers to assembly-language routines that manipulate the
 * JTAG port. uart.h refers to routines that transmit and receive
 * data over the UART.
 * 
 *  STATE PDU:
 *    1    2    3
 *  +----+----+----+--/
 *  | 00 | S1 | S2 | ...
 *  +----+----+----+--/
 *
 *  RUNTEST PDU:
 *     1    2    3    4    5    6    7   8
 *  +----+----+----+----+----+----+----+----+
 *  | 01 |       RUNCNT      | ES | RS | RC |
 *  +----+----+----+----+----+----+----+----+
 *
 *  SIRx & SDRx PDUs:
 *    1    2    3    4    5           N
 *  +----+----+----+----+------/ /------+
 *  | 0X | LENGTH  | ES |     DATA      |
 *  +----+----+----+----+------/ /------+
 *
 *  TYPE Field:
 *    00 = STATE
 *    01 = RUNTEST
 *    02 = SIRI - contains TDI for instruction register
 *    03 = SIRIO - contains TDI and requests TDO
 *                 response PDU same format
 *    04 = SDRI - contains TDI for data register
 *                response PDU same format
 *    05 = SDRIO - contains TDI and requests TDO
 *
 *  S1, S2... = sequential TAP states (any value)
 *  RUNCNT = number of clocks in run state
 *  RC (run clock) = 0 for JTAG TCK or 1 for SCK
 *  RS (run state) = RESET, IDLE, IRPAUSE or DRPAUSE
 *  ES (end state) = RESET, IDLE, IRPAUSE or DRPAUSE
 *  LENGTH = number of bits in TDI and TDO fields
 *  DATA = 0-60 bytes of data
 *
 *  TAP States are encoded as:
 *    00 = RESET	01 = IDLE	02 = DRSELECT	03 = DRCAPTURE
 *    04 = DRSHIFT	05 = DREXIT1	06 = DRPAUSE	07 = DREXIT2
 *    08 = DRUPDATE	09 = IRSELECT	0A = IRCAPTURE	0B = IRSHIFT
 *    0C = IREXIT1	0D = IRPAUSE	0E = IREXIT2	0F = IRUPDATE
 */

#include	<string.h>
#include	"NML675001.h"
#include	"DCP1-RTOS.h"
#include	"events.h"
#include	"uart.h"
#include	"jtag.h"

/*
 * download program common variables
 */
int download_pcb[32];	/* PCB with 16 word staack */

void jtag_init (void)
{
	/* initialize I/O port */
	jinitialize();
	/* register download user program to wait on flag 0 */
	#asm
	load	r0, #1
	load	r1, #port0mask
	load	r2, #_download_pcb+124
	load	r3, #_download
	swi	1
	#endasm
};

/*
 * FPGA download main program
 * Invoked first when link reset, then on every received frame.
 */
void download (void)
{
	int n,i;
	buffer *b;
	for (;;)
	{
		/* check for received data */
		b = qnext(&rxq[0]);
		if (b != 0)
		{
			/* decode frame - user data at data[1] and length is size-2 */
			n = b->size - 2;	/* get PDU length */
			switch (b->data[1])
			{
				case 0:		/* STATE PDU */
				for (i=2; i <= n; ++i)	/* GO THROUGH STATE LIST */
					jstate(b->data[i] & 0x0F);
				qfirst(&pool,b);	/* discard received buffer */
				break;
				
				case 1:		/* RUNTEST PDU - ignore clock select */
				i = getint(&b->data[2]);	/* get clock count */
				jstate(b->data[7]);	/* go to run state */
				if (b->data[7] == 0)
					jtest(i,1);	/* run test in RESET state */
				else
					jtest(i,0);	/* run test in other state */
				jstate(b->data[6]); 	/* go to end state */
				qfirst(&pool,b);	/* discard received buffer */
				break;
				
				case 2:		/* SIR PDU */
				case 3:
				i = (unsigned char)b->data[2];	/* get bit string length */
				i += (unsigned char)b->data[3] << 8;
				if ((i < 1) || (i > 480))	/* check length */
				{
					b->size = 6;		/* if wrong */
					b->data[1] = -1;	/* send error PDU */
					qlast(&txq[0],b);
					break;
				}; /* end if */								
				jstate(11);		/* go to SHIFT-IR */
				if (b->data[1] & 1)	/* check if response required */
				{				
					if (b->data[4] == 11)	/* if end sate SHIFT-IR */
						jshift(i,&b->data[5],1); /* more data to come */
					else
						jshift(i,&b->data[5],0); /* last block */
					/* TDO written to buffer by jshift() */
					qlast(&txq[0],b);	/* format response */
				} else {		/* no data in respose */
					if (b->data[4] == 11)	/* if end sate SHIFT-IR */
						jwrite(i,&b->data[5],1); /* more data to come */
					else
						jwrite(i,&b->data[5],0); /* last block */
					b->size = 6;		/* shorten frame */
					qlast(&txq[0],b);	/* and send response */
				}; /* if */
				jstate(b->data[4]);	/* go to end state */
				break;
				
				case 4:		/* SDR PDU */
				case 5:
				i = (unsigned char)b->data[2];	/* get bit string length */
				i += (unsigned char)b->data[3] << 8;
				if ((i < 1) || (i > 480))	/* check length */
				{
					b->size = 6;		/* if wrong */
					b->data[1] = -1;	/* send error PDU */
					qlast(&txq[0],b);
					break;
				}; /* end if */								
				jstate(4);		/* go to SHIFT-DR */
				if (b->data[1] & 1)	/* check if response required */
				{				
					if (b->data[4] == 4)	/* if end sate SHIFT-DR */
						jshift(i,&b->data[5],1); /* more data to come */
					else
						jshift(i,&b->data[5],0); /* last block */
					/* TDO written to buffer by jshift() */
					qlast(&txq[0],b);	/* format response */
				} else {		/* no data in respose */
					if (b->data[4] == 4)	/* if end sate SHIFT-DR */
						jwrite(i,&b->data[5],1); /* more data to come */
					else
						jwrite(i,&b->data[5],0); /* last block */
					b->size = 6;		/* shorten frame */
					qlast(&txq[0],b);	/* and send response */
				}; /* if */
				jstate(b->data[4]);	/* go to end state */
				break;
				
				default:	/* undefined */
					b->size = 3;
					b->data[1] = -1;	/* TYPE is error PDU */
					qlast(&txq[0],b);	/* send frame */
				break;
			}; /* end switch */
		}; /* end if */
		/* request one buffer of data on port 0 */
		i = semaphore(&rxf[0],1);
		/* wait for UART port 0 event */
		#asm
		swi	0
		#endasm
	};/* end for */
}; /* end download() */
 
