;
;             JTAG.AS - JTAG port handling
;
;      (c) Copyright 2004 John Stephensen  All Rights Reserved
;
; This software in source, object or binary form is licensed only for
; personal non-profit educational use in the Amateur Radio Service and
; the license is not transferrable. The software is provided as-is for
; experimental purposes and the author does not warranty its freedom
; from defects or its suitability for any specific application.
;
; This modules consists of a set of assembly language routines,
; callable from C, that manipulate the JTAG port for programming
; Xilinx FPGAs. JTAG lines should be set low normally to minimize
; power dissipation.
;
; This program ses PIO port C at address B7A01040.
;   +0: output port (non-inverting)
;   +4: input port (non-inverting)
;   +8: mode register (0=input, 1=output)
;   +C: interrupt enable register (0=disable, 1=enable)
;  +10: interrupt polarity register ()
;  +14: interrupt status register (1=interrupt)
; The following bits of PIO port C are used.
;  Bit 0: OC Driver 1
;  Bit 1: OC Driver 2
;  Bit 2: FPGA PROG_B output
;  Bit 3: FPGA TDI output
;  Bit 4: FPGA TCK output
;  Bit 5: FPGA TDO input
;  Bit 6: FPGA TMS output
;  Bit 7: FPGA DONE input
;
; port address and offsets
GPPC	EQU	0xB7A01040
GPPO	EQU	0x00
GPPI	EQU	0x04
GPPM	EQU	0x08
GPIE	EQU	0x0C
GPIP	EQU	0x10
GPIS	EQU	0x14
; bit masks
PROGB	EQU	0x0004
TDI	EQU	0x0008
TCK	EQU	0x0010
TDO	EQU	0x0020
TMS	EQU	0x0040
DONE	EQU	0x0080
; bit plus clock combinations
RESET	EQU	TCK
TCK0	EQU	PROGB
TCK1	EQU	PROGB|TCK
TMS0	EQU	PROGB
TMS1	EQU	PROGB|TMS
TDI0	EQU	PROGB
TDI1	EQU	PROGB|TDI
;
; JTAG TAP state
;
	PSECT	bss, class=RAM
TSTATE:	DS	1
;
; external entry points
;
	PSECT	text, class=CODE, reloc=4
	GLOBAL	_jinitialize
	GLOBAL	_jtest
	GLOBAL	_jstate
	GLOBAL	_jselect
	GLOBAL	_jshift
	GLOBAL	_jwrite
;
; JTAG PIO Port and FPGA Initialization
;
_jinitialize:
	LOAD	R0, #GPPC	; port base address
	MOV	R12, #0x0000	; disable interrupts
	STRH	R12, [R0, #GPIE]
	MOV	R12, #0x005F	; bits 5 and 7 are inputs
	STRH	R12, [R0, #GPPM]
	MOV	R12, #RESET	; reset FPGA
	STRH	R12, [R0, #GPPO]
	MOV	R12, #TCK1	; release reset
	STRH	R12, [R0, #GPPO]
	LOAD	R0, #150000	; wait 10 ms
jinit0:
	SUBS	R0, R0, #1	; 4 clocks per loop
	BNE	jinit0		; is 15 counts per us
	LOAD	R0, #TSTATE
	MOV	R1, #0		; reset TAP state
	STRB	R1, [R0]
	MOV	R0, #0
	B	_jstate		; return via jstate
;
; JTAG test routine
;
; Called with:
;	R0 = count (microseconds)
;	R1 = zero if TMS = 0 or non-zero if TMS = 1
; Uses:
;	R2 = port address
;	R12 = temp.
;
; Sends R0 TCK clocks at one clock per microsecond.
; 61 clocks per loop for +1.7% at 60 MHz.
;
_jtest:
	CMP	R0, #0		; done if zero clocks
	MOVEQ	PC, LR
	LOAD	R2, #GPPC	; JTAG port address
jtest0:
	CMP	R1, #0		; select TMS state
	MOVEQ	R12, #TMS0
	MOVNE	R12, #TMS1
	STRH	R12, [R2, #GPPO] ; TCK = 0
	ORR	R12, R12, #TCK1
	STRH	R12, [R2, #GPPO] ; TCK = 1
	MOV	R12, #12	; wait 48 clocks
jtest1:
	SUBS	R12, R12, #1	; at 4 clocks per loop
	BNE	jtest1
	SUBS	R0, R0, #1
	BNE	jtest0
	MOV	PC, LR		; return
;
; JTAG TAP State Transition Table
;
; Row labels are old state and column labels are new state.
;
JTSTT:
;          |<------DATA REGISTER------>|<---INSTRUCTION REGISTER--->|
;   RST IDL SEL CAP SFT EX1 PSE EX2 UPD SEL CAP SFT EX1 PSE  EX2 UPD		
 DB 1FH,00H,02H,02H,02H,0AH,0AH,2AH,1AH,06H,06H,06H,16H,56H,036H,1EH ; RESET
 DB 07H,00H,01H,01H,01H,05H,05H,15H,0DH,03H,03H,03H,0BH,0BH,02BH,1BH ; IDLE
 DB 1FH,03H,00H,00H,00H,02H,02H,0AH,06H,01H,01H,01H,05H,05H,015H,0BH ; SELECT-DR
 DB 1FH,03H,07H,00H,00H,01H,01H,05H,03H,0FH,0FH,0FH,2FH,2FH,0AFH,6FH ; CAPTURE-DR
 DB 1FH,03H,07H,07H,00H,01H,01H,05H,03H,0FH,0FH,0FH,2FH,2FH,0AFH,6FH ; SHIFT-DR
 DB 1FH,01H,03H,03H,03H,00H,00H,02H,01H,07H,07H,07H,17H,17H,057H,37H ; EXIT1-DR
 DB 1FH,03H,07H,07H,01H,17H,17H,01H,03H,0FH,0FH,0FH,2FH,2FH,0AFH,6FH ; PAUSE-DR
 DB 1FH,01H,03H,03H,00H,0BH,0BH,00H,01H,07H,07H,07H,17H,17H,057H,37H ; EXIT2-DR
 DB 1FH,00H,01H,01H,01H,05H,05H,15H,00H,03H,03H,03H,0BH,0BH,02BH,1BH ; UPDATE-DR
 DB 1FH,01H,05H,05H,05H,15H,15H,55H,35H,00H,00H,00H,02H,02H,00AH,06H ; SELECT-IR
 DB 1FH,03H,07H,07H,07H,17H,17H,57H,37H,0FH,00H,00H,01H,01H,005H,03H ; CAPTURE-IR
 DB 1FH,03H,07H,07H,07H,17H,17H,57H,37H,0FH,0FH,00H,01H,01H,005H,03H ; SHIFT-IR
 DB 1FH,01H,03H,03H,03H,0BH,0BH,2BH,1BH,07H,07H,07H,00H,00H,002H,01H ; EXIT1-IR
 DB 1FH,03H,07H,07H,07H,17H,17H,57H,37H,0FH,0FH,0FH,2FH,2FH,001H,03H ; PAUSE-IR
 DB 1FH,01H,03H,03H,03H,0BH,0BH,2BH,1BH,07H,07H,00H,17H,17H,000H,01H ; EXIT2-IR
 DB 1FH,00H,01H,01H,01H,05H,05H,15H,0DH,03H,03H,03H,0BH,0BH,02BH,00H ; UPDATE-IR
 DB   5,  1,  2,  3,  4,  4,  5,  6,  5,  3,  4,  5,  5,  6,   7,  6 ; RESET
 DB   3,  0,  1,  2,  3,  3,  4,  5,  4,  2,  3,  4,  4,  5,   6,  5 ; IDLE
 DB   5,  3,  0,  1,  2,  2,  3,  4,  3,  1,  2,  3,  3,  4,   5,  4 ; SELECT-DR
 DB   5,  3,  3,  0,  1,  1,  2,  3,  2,  4,  5,  6,  6,  7,   8,  7 ; CAPTURE-DR
 DB   5,  3,  3,  4,  0,  1,  2,  3,  2,  4,  5,  6,  6,  7,   8,  7 ; SHIFT-DR
 DB   5,  2,  2,  3,  4,  0,  1,  2,  1,  3,  4,  5,  5,  6,   7,  6 ; EXIT1-DR
 DB   5,  3,  3,  4,  2,  5,  6,  1,  2,  4,  5,  6,  6,  7,   8,  7 ; PAUSE-DR
 DB   5,  2,  2,  3,  1,  4,  5,  0,  1,  3,  4,  5,  5,  6,   7,  6 ; EXIT2-DR
 DB   5,  1,  1,  2,  3,  3,  4,  5,  0,  2,  3,  4,  4,  5,   6,  5 ; UPDATE-DR
 DB   5,  2,  3,  4,  5,  5,  6,  7,  6,  0,  1,  2,  2,  3,   4,  3 ; SELECT-IR
 DB   5,  3,  3,  4,  5,  5,  6,  7,  6,  4,  0,  1,  1,  2,   3,  2 ; CAPTURE-IR
 DB   5,  3,  3,  4,  5,  5,  6,  7,  6,  4,  5,  0,  1,  2,   3,  2 ; SHIFT-IR
 DB   5,  2,  2,  3,  4,  4,  5,  6,  5,  3,  4,  5,  0,  1,   2,  1 ; EXIT1-IR
 DB   5,  3,  3,  4,  5,  5,  6,  7,  6,  4,  5,  6,  6,  7,   1,  2 ; PAUSE-IR
 DB   5,  2,  2,  3,  4,  4,  5,  6,  5,  3,  4,  1,  5,  6,   0,  1 ; EXIT2-IR
 DB   5,  1,  1,  2,  3,  3,  4,  5,  4,  2,  3,  4,  4,  5,   6,  0 ; UPDATE-IR
;   RST IDL SEL CAP SFT EX1 PSE EX2 UPD SEL CAP SFT EX1 PSE  EX2 UPD		
;          |<------DATA REGISTER------>|<---INSTRUCTION REGISTER--->|
;
; JTAG State Transition Routine
;
; Called with:
;	R0 = next state
; Uses:
;	R1 = state transition table index
;	R2 = state transition table base
;
; Looks up TMS sequence and length in JTAG state transition
; table. Lower 5 bits are TMS bit sequence (rightmost bit first)
; and upper 3 bits are length. Table is index with LS nibble
; containing new state and MS nibble containing old state.
; Drops through to jselect.
;
_jstate:
	AND	R0, R0, #0x0F	; mask state bits
	LOAD	R2, #TSTATE	; TAP state variable
	LDRB	R1, [R2]	; get old value
	AND	R1, R1, #0x0F	; mask state bits
	STRB	R0, [R2]	; store new value
	ORR	R0, R0, R1, LSL #4 ; create index
	LOAD	R2, #JTSTT	; JTAG state transition table
	LDRB	R1, [R2, R0]	; get entry
	ADD	R0, R0, #256	; offset to length
	LDRB	R0, [R2, R0]	; get length
;
; JTAG Select 
;
; Called with:
;	R0 = bit count
;	R1 = TMS bit pattern (right justified)
; Uses:
;	R2 = I/O port address
;	R12 = temp
;
; Clocks 1-8 bit TMS patterns while TDI remains zero.
; All zeros or all ones patterns can be repeated.
; 12 clock loop provides 5 MHz shift rate at 60 MHz. 
;
_jselect:
	CMP	R0, #0		; done if zero
	MOVEQ	PC, LR		; return
	LOAD	R2, #GPPC	; port base address
jselect0:
	MOVS	R1, R1, ROR #1	; TMS bit to carry flag
	MOVCC	R12, #TMS0	; send zero
	MOVCS	R12, #TMS1	; or send one
	STRH	R12, [R2, #GPPO]
	ORR	R12, R12, #TCK	; clock active edge
	STRH	R12, [R2, #GPPO]
	SUBS	R0, R0, #1	; decrement counter
	BNE	jselect0	; loop until all data sent
	MOV	PC, LR		; return to caller
;
; JTAG Shift
;
; Called with:
;	R0 = number of bits to transfer
;	R1 = address of right justified TDI bits
;	R2 = flags
;	TAP in SHIFT-xR state
; Uses:
;	R3 = TDI and TDO bits
;	R4 = bit count
;	R% = temp
;	LR = I/O port address
;	R12 = JTAG bit pattern
; Returns:
;	R1 was address of right justified TDO bits
;	TAP in EXIT1-xR state if flags zero
;	TAP in SHIFT-xR state otherwise
;
; Shifts data out of and in to instruction or data register.
; 30 clocks per loop for 2 MBPS at 60 MHz.
;
_jshift:
	CMP	R0, #0		; check length
	CMPNE	R1, #0		; and address
	MOVEQ	PC, LR		; return if zero
	STMFD	SP!, {R4,R5,LR}	; save return address
	MOV	R4, #0		; reset bit counter
	LOAD	LR, #GPPC	; port base address
	MOV	R12, #TMS0	; TMS=0
jshift0:
	TST	R4, #7		; if count mod 8 = 0
	LDREQB	R3, [R1]	; then load byte
	BIC	R12, R12, #TCK	; clock trailing edge
	STRH	R12, [LR, #GPPO] ; shifts out data
	MOVS	R3, R3, ROR #1	; TDI bit to carry flag
	MOVCC	R12, #TDI0	; and send zero
	MOVCS	R12, #TDI1	; or send one
	ADD	R4, #1		; increment bit counter
	CMP	R4, R0		; if last bit
	CMPEQ	R2, #0		; check more flag
	ORREQ	R12, R12, #TMS	; set TMS if last bit
	STRH	R12, [LR, #GPPO] ; set TDI and TMS
	LDRH	R5, [LR, #GPPI] ; sample JTAG port
	ANDS	R5, #TDO	; check TDO bit
	BICEQ	R3, R3, #0x80	; and add zero
	ORRNE	R3, R3, #0x80	; or add one
	ORR	R12, R12, #TCK	; clock leading edge
	STRH	R12, [LR, #GPPO] ; to change state
	TST	R4, #7		; if count mod 8 = 0
	STREQB	R3, [R1]	; then store byte
	ADDEQ	R1, R1, #1	; and increment address
	CMP	R4, R0		; compare count against limit
	BNE	jshift0		; and loop until all bits sent
	RSBS	R0, R0, #0	; calculate residual offset (0-7)
	AND	R0, R0, #7	; (negative length mod 8)
	MOV	R3, R3, ROR R0	; adjust TDO and
	STRNEB	R3, [R1]	; save last byte if offset not 0
	CMP	R2, #0		; check for last block
	LDMNEFD	SP!, {R4,R5,PC}	; and return if not
	LOAD	R12, #TSTATE	; show state transition
	LDRB	R0, [R12]
	ADD	R0, R0, #1	; by incrementing state
	STRB	R0, [R12]
	LDMFD	SP!, {R4,R5,PC}	; return
;
; JTAG Data Byte Write
;
; Called with:
;	R0 = number of bits to transfer
;	R1 = address of data bytes to send
;	R2 = flags
;	TAP in SHIFT-xR state
; Uses:
;	R3 = data byte
;	LR = I/O port address
;	R12 = temp
; Returns with:
;	TAP in EXIT1-xR state if last block
;	TAP in SHIFT-xR state otherwise
;
; Data is written MSB first. 8 clocks per bit for
; 7.5 MBPS at 60 MHz. 
;
_jwrite:
	CMP	R0, #0		; check parameters
	CMPNE	R1, #0
	MOVEQ	PC, LR		; and return if any zero
	STMFD	SP!, {LR}	; save return address
	LOAD	LR, #GPPC	; port base address
jwrite0:
	LDRB	R3, [R1]	; load byte
	ADD	R1, R1, #1	; and increment address
	;
	TST	R3, #0x01	; test bit 0
	MOVEQ	R12, #TDI0	; set data bit to zero or
	MOVNE	R12, #TDI1	; set data bit to one
	SUBS	R0, R0, #1	; decrement bit count?
	BEQ	jwrite1		; done if last bit
	STRH	R12, [LR, #GPPO]
	ORR	R12, R12, #TCK	; do clock transition
	STRH	R12, [LR, #GPPO]
	;
	TST	R3, #0x02	; test bit 1
	MOVEQ	R12, #TDI0	; set data bit to zero or
	MOVNE	R12, #TDI1	; set data bit to one
	SUBS	R0, R0, #1	; decrement bit count?
	BEQ	jwrite1		; done if last bit
	STRH	R12, [LR, #GPPO]
	ORR	R12, R12, #TCK	; do clock transition
	STRH	R12, [LR, #GPPO]
	;
	TST	R3, #0x04	; test bit 2
	MOVEQ	R12, #TDI0	; set data bit to zero or
	MOVNE	R12, #TDI1	; set data bit to one
	SUBS	R0, R0, #1	; decrement bit count?
	BEQ	jwrite1		; done if last bit
	STRH	R12, [LR, #GPPO]
	ORR	R12, R12, #TCK	; do clock transition
	STRH	R12, [LR, #GPPO]
	;
	TST	R3, #0x08	; test bit 3
	MOVEQ	R12, #TDI0	; set data bit to zero or
	MOVNE	R12, #TDI1	; set data bit to one
	SUBS	R0, R0, #1	; decrement bit count?
	BEQ	jwrite1		; done if last bit
	STRH	R12, [LR, #GPPO]
	ORR	R12, R12, #TCK	; do clock transition
	STRH	R12, [LR, #GPPO]
	;
	TST	R3, #0x10	; test bit 4
	MOVEQ	R12, #TDI0	; set data bit to zero or
	MOVNE	R12, #TDI1	; set data bit to one
	SUBS	R0, R0, #1	; decrement bit count?
	BEQ	jwrite1		; done if last bit
	STRH	R12, [LR, #GPPO]
	ORR	R12, R12, #TCK	; do clock transition
	STRH	R12, [LR, #GPPO]
	;
	TST	R3, #0x20	; test bit 5
	MOVEQ	R12, #TDI0	; set data bit to zero or
	MOVNE	R12, #TDI1	; set data bit to one
	SUBS	R0, R0, #1	; decrement bit count?
	BEQ	jwrite1		; done if last bit
	STRH	R12, [LR, #GPPO]
	ORR	R12, R12, #TCK	; do clock transition
	STRH	R12, [LR, #GPPO]
	;
	TST	R3, #0x40	; test bit 6
	MOVEQ	R12, #TDI0	; set data bit to zero or
	MOVNE	R12, #TDI1	; set data bit to one
	SUBS	R0, R0, #1	; decrement bit count?
	BEQ	jwrite1		; done if last bit
	STRH	R12, [LR, #GPPO]
	ORR	R12, R12, #TCK	; do clock transition
	STRH	R12, [LR, #GPPO]
	;
	TST	R3, #0x80	; test bit 7
	MOVEQ	R12, #TDI0	; set data bit to zero or
	MOVNE	R12, #TDI1	; set data bit to one
	SUBS	R0, R0, #1	; decrement bit count?
	BEQ	jwrite1		; done if last bit
	STRH	R12, [LR, #GPPO]
	ORR	R12, R12, #TCK	; do clock transition
	STRH	R12, [LR, #GPPO]
	B	jwrite0
jwrite1:
	CMP	R2, #0		; check more flag
	ORREQ	R12, R12, #TMS	; set TMS if last bit
	STRH	R12, [LR, #GPPO]
	ORR	R12, R12, #TCK	; do clock transition
	STRH	R12, [LR, #GPPO]
	CMP	R2, #0		; check for last block
	LDMNEFD	SP!, {PC}	; and return if not
	LOAD	R12, #TSTATE	; show state transition
	LDRB	R0, [R12]
	ADD	R0, R0, #1	; by incrementing state
	STRB	R0, [R12]
	LDMFD	SP!, {PC}	; return
;

