;
;             QUEUES.AS - Linked list queue handling
;
;      (c) Copyright 2004 John Stephensen  All Rights Reserved
;
; This software in source, object or binary form is licensed only for
; personal non-profit educational use in the Amateur Radio Service and
; the license is not transferrable. The software is provided as-is for
; experimental purposes and the author does not warranty its freedom
; from defects or its suitability for any specific application.
;
; This modules consists of a set of assembly language routines,
; callable from C, that manipulte singly-linked lists (queues).
; Since they do not use R4-R11, these registers do not have to
; be saved on the stack as in a C-language implementation.
;
; Linked lists are controlled by a structure containing a 32-bit
; address pointing to the first entry in the list (head) followed by
; a 32-bit address pointing to the last entry in the list (tail).
;
; List entries start with a 32-bit address (link field) that points
; to the next entry on this list. The link in the last entry is set
; to NUL. Empty lists have the head and tail pointers set to NUL.
;
; The structure following the link field is not used by this library
; of functions. If the entry is a buffer, there is a 32-bit size
; field following the link field.
;
; The following functions are defined for manipulating lists:
;
;	qnew - initialize list with no entries
;	qfill - initialize list with multiple entries
;	qfirst - add item to head of list
;	qmfirst - add multiple entries to head of list
;	qlast - add item to tail end of list
;	qmlast - add multiple entries to end of list
;	qcheck - return address of first item on list
;	qnext - remove one item from head of list
;	qcount - return number of entries in list
;
; The qfirst, qlast, qnext, qmfirst, qmlast and qcount functions contain
; critical sections that are protected by disabling interrupts. The
; qcheck() function needs no protection. The qnew and qfill functions
; should only be called with interrupts disabled and they do not
; re-enable interrupts. Do not use qnew to flush a buffer. qmfirst and
; qmlast can be used to remove all buffers from one list and add them
; to another.
; 
; There are also functions to manipulate the contents of a buffer:
;
;	parity - calculate even parity for each bit of each byte
;	putint - place 32-bit integer into buffer in any alignment
;	getint - retreive 32-bit integer from buffer in any alignment
;	putzero - zero the next 4 bytes
;
; There is also one function to manipulate a semaphore. The semaphore
; can be reset to zero, incremented by any amount, or decremented if
; the resulting value is not negative. Interrupts are disabled during
; the critical section so that user and sytem programs can use semaphores
; to communicate. Do not use this function with interrupts disabled as it
; will re-enable them. Semaphores can be reset by writing zero to them
; with interrupts disabled.
;
	PSECT	text, class=CODE, reloc=4
	GLOBAL	_qnew
	GLOBAL	_qfirst
	GLOBAL	_qlast
	GLOBAL	_qnext
	GLOBAL	_qcheck
	GLOBAL	_qfill
	GLOBAL	_qmfirst
	GLOBAL	_qmlast
	GLOBAL	_qcount
	GLOBAL	_parity
	GLOBAL	_NUL
	GLOBAL	_putint
	GLOBAL	_getint
	GLOBAL	_putzero
	GLOBAL	_semaphore
;
_NUL	EQU	0
;
; Initialize queue control block to zero
;
; qnew(void *list);
;
; Called with:
;	R0 = list pointer
; Uses:
;	R12 = temporary register
;
_qnew:
	CMP	R0, #0		; check all parameters for NUL
	MOVEQ	PC, LR		; return if any parameter NUL
	MOV	R12, #0		; NUL = 0
	STR	R12, [R0]	; head pointer = NUL
	STR	R12, [R0, #4]	; tail pointer = NUL
	MOV	PC, LR		; return
;
; Add entry to head of list
;
; qfirst(void *list, void *entry);
;
; Called with:
;	R0 = list pointer
;	R1 = entry to be added
; Uses:
;	R2 = modified status register
;	R3 = saved status register
;	R12 = head pointer
;
_qfirst:
	CMP	R0, #0		; check all parameters for NUL
	CMPNE	R1, #0
	MOVEQ	PC, LR		; return if any parameter NUL
	MRS	R3, CPSR	; save flags
	ORR	R2, R3, #0xC0	; set interrupt mask
	MSR	CPSR_c, R2	; set flags
	LDR	R12, [R0]	; old head pointer
	STR	R12, [R1]	; is first entry link
	STR	R1, [R0]	; set new head pointer
	CMP	R12, #0		; check for empty list
	STREQ	R1, [R0, #4]	; and set tail pointer
	MSR	CPSR_c, R3	; restore interrupt mask
	MOV	PC, LR		; return
;
; Add entry to tail of list
;
; qlast(void *list, void *entry);
;
; Called with:
;	R0 = list pointer
;	R1 = entry to be added
; Uses:
;	R2 = modified status register
;	R3 = saved status register
;	R12 = tail pointer
;
_qlast:
	CMP	R0, #0		; check all parameters for NUL
	CMPNE	R1, #0
	MOVEQ	PC, LR		; return if any parameter NUL
	MOV	R12, #0		; 0 = NUL
	MRS	R3, CPSR	; save flags
	ORR	R2, R3, #0xC0	; set interrupt mask
	MSR	CPSR_c, R2	; set flags
	STR	R12, [R1]	; set last entry link
	LDR	R12, [R0, #4]	; old tail pointer
	STR	R1, [R0, #4]	; new tail pointer
	CMP	R12, #0		; check for empty list
	STREQ	R1, [R0]	; set head pointer or
	STRNE	R1, [R12]	; previous entry link
	MSR	CPSR_c, R3	; restore interrupt mask
	MOV	PC, LR		; return
;
; Remove entry from head of list
;
; *void qnext(void *list);
;
; Called with:
;	R0 = list pointer
; Returns:
;	R0 = address of entry that was removed or NUL
; Uses:
;	R1 = removed entry link
;	R2 = modified status register
;	R3 = saved status register
;	R12 = list pointer
;
_qnext:
	CMP	R0, #0		; check all parameters for NUL
	MOVEQ	PC, LR		; return if any parameter NUL
	MOV	R12, R0		; make space for result
	MRS	R3, CPSR	; save flags
	ORR	R2, R3, #0xC0	; set interrupt mask
	MSR	CPSR_c, R2	; set flags
	LDR	R0, [R12]	; list head
	CMP	R0, #0		; check for NUL
	BEQ	qnext0		; return if empty
	LDR	R1, [R0]	; address of next entry
	STR	R1, [R12]	; new list head
	CMP	R1, #0		; if NUL
	STREQ	R1, [R12, #4]	; set tail pointer
qnext0:
	MSR	CPSR_c, R3	; restore interrupt mask
	MOV	PC, LR		; return
;
; Returns address of entry at head of list
;
; *void qcheck(void *list);
;
; Called with:
;	R0 = list pointer
; Returns:
;	R0 = address of entry that was removed or NUL
; Uses:
;	R12 = list pointer
;
_qcheck:
	CMP	R0, #0		; check all parameters for NUL
	MOVEQ	PC, LR		; return if any parameter NUL
	MOV	R12, R0		; make space for result
	LDR	R0, [R12]	; list head
	MOV	PC, LR		; return
;
; Populate list with empty buffers
;
; qfill(void *list, void *ram, int number, int size);
;
; Called with:
;	R0 = list pointer
;	R1 = pointer to block of RAM (word aligned)
;	R2 = number of buffers to create
;	R3 = size of each buffer in words (including link & size)
; Uses:
;	R12 = previous link field value
;	LR = zero
;	4 bytes of stack space
;
_qfill:
	CMP	R0, #0		; check all parameters for NUL
	CMPNE	R1, #0
	CMPNE	R2, #0
	CMPNE	R3, #2		; must have link & size plus data
	MOVLE	PC, LR		; return if any parameter NUL
	STMFD	SP!,{LR}	; save return address on stack
	MOV	R3, R3, LSL #2	; words -> bytes
	SUB	R3, R3, #8	; data field length is less link and size
	MOV	LR, #0		; need zero to clear size field
	MOV	R12, #0		; NUL link
	STR	R1, [R0, #4]	; first block at tail end of list
qfill0:
	STR	R12, [R1]	; set link field
	STR	LR, [R1, #4]	; clear size field
	MOV	R12, R1		; save link address
	ADD	R1, R1, R3	; go to next buffer
	ADD	R1, R1, #8	; add space for link and size fields
	SUBS	R2, R2, #1	; decrement buffer number
	BNE	qfill0		; loop if more buffers
	STR	R12, [R0]	; last block at head of list
	LDMFD	SP!,{PC}	; return
;
; Move source list to front of destination list
; qmfirst(void *dst, *src);
; Called with:
;	R0 = destination list pointer
;	R1 = source list pointer
; Uses:
;	R2 = destination head pointer
;	R3 = source head & tail pointers
;	R12 = saved flags
;
_qmfirst:
	CMP	R0, #0		; check for NUL pointers
	CMPNE	R1, #0
	MOVEQ	PC, LR		; return if either NUL
	MRS	R12, CPSR	; save flags
	ORR	R3, R12, #0xC0	; set interrupt mask
	MSR	CPSR_c, R3	; set flags
	LDR	R3, [R1, #4]	; get source tail pointer
	CMP	R3, #0		; check for NUL
	BEQ	qmfirst0	; done if source list empty
	LDR	R2, [R0]	; get destination head pointer
	CMP	R2, #0		; check for empty list
	STREQ	R3, [R0, #4]	; if so, dst = src tail pointer
	STRNE	R2, [R3]	; if not, last src entry link = dst head ptr
	LDR	R3, [R1]	; get source head pointer
	STR	R3, [R0]	; set destination head pointer
	MOV	R0, #0		; zero source pointers
	STR	R0, [R1]
	STR	R0, [R1, #4]
qmfirst0:
	MSR	CPSR_c, R12	; restore flags
	MOV	PC, LR		; return
;
; Move source list to end of destination list
; qmfirst(void *dst, *src);
; Called with:
;	R0 = destination list pointer
;	R1 = source list pointer
; Uses:
;	R2 = destination head pointer
;	R3 = source head & tail pointers
;	R12 = saved flags
;
_qmlast:
	CMP	R0, #0		; check for NUL pointers
	CMPNE	R1, #0
	MOVEQ	PC, LR		; return if either NUL
	MRS	R12, CPSR	; save flags
	ORR	R3, R12, #0xC0	; set interrupt mask
	MSR	CPSR_c, R3	; set flags
	LDR	R3, [R1]	; get source head pointer
	CMP	R3, #0		; check for NUL
	BEQ	qmlast0		; done if source list empty
	LDR	R2, [R0, #4]	; get destination tail pointer
	CMP	R2, #0		; check for empty list
	STREQ	R3, [R0]	; if so, dst = src head pointer
	STRNE	R3, [R2]	; if not, last dst entry link = src head ptr
	LDR	R3, [R1, #4]	; get source tail pointer
	STR	R3, [R0, #4]	; set destination tail pointer
	MOV	R0, #0		; zero source pointers
	STR	R0, [R1]
	STR	R0, [R1, #4]
qmlast0:
	MSR	CPSR_c, R12	; restore flags
	MOV	PC, LR		; return
;
; Returns number of entries in list
; *void qcount(void *list);
; Called with:
;	R0 = list pointer
; Returns:
;	R0 = number of entries
; Uses:
;	R2 = modified status register
;	R3 = saved status register
;	R12 = list pointer
;
_qcount:
	CMP	R0, #0		; check all parameters for NUL
	MOVEQ	PC, LR		; return if any parameter NUL
	MOV	R12, R0		; set list pointer
	MVN	R0, #0		; set entry count to -1
	MRS	R3, CPSR	; save flags
	ORR	R2, R3, #0xC0	; set interrupt mask
	MSR	CPSR_c, R2	; set flags
count0:
	ADD	R0, R0, #1	; increment count
	LDR	R12, [R12]	; get next link
	CMP	R12, #0		; check for end of list
	BNE	count0		; loop if more entries
	MSR	CPSR_c, R3	; restore interrupt mask
	MOV	PC, LR		; return if end
;
; Calculate even parity
;
; char = parity(char *data, int size)
;
; Called with:
;	R0 = address of first character
;	R1 = size of buffer
; Uses:
;	R2 = checksum
;	R12 = buffer
; Returns:
;	R0 = checksum
;
_parity:
	CMP	R0, #0
	CMPNE	R1, #0
	MOVEQ	PC, LR		; return if nothing
	MOV	R2, #0		; initialize checksum
parity0:
	LDRB	R12, [R0]	; get character
	EOR	R2, R2, R12	; and add to sum
	ADD	R0, R0, #1	; increment pointer
	SUBS	R1, R1, #1	; decrement counter
	BNE	parity0		; loop until done
	ANDS	R0, R2, #0xFF	; LSB of sum
	MOV	PC, LR		; return
;
; putint(char *data, int value)
;
; Called with:
;	R0 = address of first byte (LSB)
;	R1 = 32-bit integer value
; Uses:
;	R12 = temp
;
_putint:
	CMP	R0, #0
	MOVEQ	PC, LR		; return if nothing
putint0:
	STRB	R1, [R0]	; 1st byte
	MOV	R12, R1, LSR #8		; shift
	STRB	R12, [R0, #1]	; 2nd byte
	MOV	R12, R1, LSR #16	; shift
	STRB	R12, [R0, #2]	; 3rd byte
	MOV	R12, R1, LSR #24	; shift
	STRB	R12, [R0, #3]	; 4th byte
	MOV	PC, LR		; done
;
; int = getint(char *data)
;
; Called with:
;	R0 = address of first byte (LSB)
; Returns:
;	R0 = 32-bit integer value
; Uses:
;	R0-R3, R12
;
_getint:
	CMP	R0, #0
	MOVEQ	PC, LR		; return if nothing
getint0:
	LDRB	R1, [R0]	; 1st byte
	LDRB	R2, [R0, #1]	; 2nd byte
	LDRB	R3, [R0, #2]	; 3rd byte
	LDRB	R12, [R0, #3]	; 4th byte
	ORR	R1, R1, R2, LSL #8	; shift
	ORR	R1, R1, R3, LSL #16	; shift
	ORR	R0, R1, R12, LSL #24	; shift
	MOV	PC, LR		; done
;
; putzero(char *data)
;
; Called with:
;	R0 = address of first byte (LSB)
; Uses:
;	R12
;
_putzero:
	CMP	R0, #0
	MOVEQ	PC, LR		; return if nothing
	MOV	R12, #0
	STRB	R12, [R0]	; 1st byte
	STRB	R12, [R0, #1]	; 2nd byte
	STRB	R12, [R0, #2]	; 3rd byte
	STRB	R12, [R0, #3]	; 4th byte
	MOV	PC, LR		; done
;
; Semaphore Manipulation Routine
;
; Called with:
;	R0 = address of semaphore
;	R1 = 0 for reset
;	     + to increment
;	     - to decrement
; Returns:
;	R0 = previous value
; Uses:
;	R2 = old value
;	R3 = new value
;	R12 = saved flags
;
_semaphore:
	CMP	R0, #0		; check address
	MOVEQ	PC, LR		; return if NUL
	MRS	R12, CPSR	; save flags
	ORR	R3, R12, #0xC0	; set interrupt mask
	MSR	CPSR_c, R3	; set flags
	LDR	R2, [R0]	; get value
	ADDS	R3, R2, R1	; add/subtract
	STRGE	R3, [R0]	; save if > 0
	CMP	R1, #0		; if reset
	STREQ	R1, [R0]	; clear semaphore
	MSR	CPSR_c, R12	; resatore old flags
	MOV	R0, R2		; get old value
	MOV	PC, LR		; return
;


