/*
 *			ARM7 UART Driver
 *
 *      (c) Copyright 2004 John B. Stephensen  All Rights Reserved
 *
 * This software in source, object or binary form is licensed only for
 * personal non-profit educational use in the Amateur Radio Service and
 * the license is not transferrable. The software is provided as-is for
 * experimental purposes and the author does not warranty its freedom
 * from defects or its suitability for any specific application.
 *
 * The communications link may use RS-232 or RS-485 signalling.
 * The RS-485 transceiver is enabled if valid RS-232 levels are
 * not present on the serial interface. Operation is half-duplex.
 * Communications is always initiated by a primary station that
 * transmits data to a selected secondary station. The primary
 * station may also poll each secondary station for data.
 *
 * This driver resides on a secondary station and implements a
 * communications protocol between 1 to 15 secondary stations
 * and a single primary station.  It is written in HI-TECH
 * ARM-C for RTOS and consists of two modules. The protocol
 * implementation executes in user mode. The frame processing
 * occurs in the interrupt service routine. Note that constants
 * are avoided to maximize the efficiency of the generated ARM7
 * assembly language code.
 *
 * The frame is transmitted using 8-bit characters with even parity
 * and one stop bit. Three special characters are used to acheive
 * 8-bit transparency:
 *
 *	7F (127) is the abort character.
 *	7E (126) is the flag character.
 *	7D (125) is the escape character.
 *
 * The characters 7E and 7F never occur within a frame. If they
 * occur in the data to be transmitted, 7F is encoded as 7D 5F and
 * 7E is encoded as 7D 5E. 7D is encoded as 7D 5D. On reception,
 * any occurrance of 7D results in its replacement by the following
 * character exclusive OR'ed with 20.
 *
 * A normal frame consists of the flag character followed by any
 * number encoded data characters and ending with a flag character.
 * If a transmission error occurs, an abort character is inserted
 * into the data field. The receiver then discards the frame. Two
 * or more flag characters may occur between frames. 
 *
 * The frame consists of a user data field preceded by a one byte
 * header field and followed by a one byte parity field. The last
 * byte of frame is even parity over each bit of preceding bytes.
 * The user data field may be 0 to 64 bytes in length.
 *
 *	    0       1                         N      N+1
 *	+--------+--------------/ /--------------+--------+
 *	| HEADER |        USER DATA FIELD        | PARITY |
 *	+--------+--------------/ /--------------+--------+
 *
 * The user data field is generated and decoded by the user of the
 * protocol. The header and parity bytes are added by the protocol
 * implementation described here on transmission and decoded by it
 * on reception. 
 *
 * The header byte contains three subfields, a 4-bit address field
 * (ADDR), a 1-bit poll/acknowledge field (PA), a 1-bit sequence
 * number field (SN) and a 2-bit port number (PORT). The address
 * identifies the secondary station and the port identifies one
 * of four communication endpoints within a station. The PA bit
 * is set to one on acknowlegement frames sent to the primary
 * station. It is also set to one on polling frames sent to the
 * secondary stations. It is set to zero in data frames. The
 * sequence bit is used to detect receipt of duplicate frames.
 * only one frame is outstanding at any time on each port. Port
 * 3 is used for control. Ports 0-2 are used for general-purpose
 * communication. Address 15 is reserved.
 *
 *	  7    6    5    4    3    2    1    0
 *	+----+----+----+----+----+----+----+----+
 * 	|        ADDR       | PA | SN |  PORT   |
 *	+----+----+----+----+----+----+----+----+
 *
 * The UART driver maintains the variables txs[port] and rxs[port]
 * with the last transmitted and received sequence numbers. These
 * determine the sequence numbers in transmitted data and
 * acknowledgement frames. The variables rxf[port] are used for
 * receiver flow control. A value of zero indicates that no new
 * data is to be accepted. Positive values indicate the number of
 * new data frames to be accepted and this number is decremented for
 * each new frame received.
 *
 * When a data frame is received, the received sequence number
 * is compared to the next expected sequence number (rxs[port]). 
 * If they are not equal, the frame is a duplicate, is acknowledged
 * with its sequence number and discarded. If the sequence numbers 
 * are equal and the flow control variable is zero, the frame is
 * discarded and acknowledged with an unequal sequence number. If
 * the flow control variable is positive, the next expected sequence
 * number is incremented, the frame is acknowledged with its
 * sequence number and the frame is posted to the receive queue for
 * the addressed port.
 *
 * When a polling frame is received, the transmit queue is checked.
 * if there is no data available, an empty data frame is returned with
 * a sequence number not equal to the sequence number in the polling
 * frame. If data is available, the received sequence number is compared
 * with the next expected sequence number (txs[port]). If the two are
 * equal, the current data frame is used. If they are not equal, the
 * current data frame is discarded as the polling frame acknowledges
 * its receipt. If a data frame is still available, it is returned with
 * a sequence number equal to the sequence number of the polling frame.  
 *
 * When a frame addressed to port 3 with the PA bit set to zero and the SN  
 * bit set to zero is received, the secondary station resets the link state
 * for all ports and returns a frame with an identical header.
 *
 * When a frame addressed to port 3 with the PA bit set to zero and the SN  
 * bit set to one is received, the secondary station returns a frame with 
 * an identical header and a 32-byte data field containing:
 *
 *        Byte 1: Zero
 *        Byte 2: Current receive buffer pool size.
 *        Byte 3: Current transmit buffer pool size.
 *     Bytes 4-7: Number of byte parity errors since last reset.
 *    Bytes 8-11: Number of frame parity errors since last reset.
 *   Bytes 10-15: Number of byte overrun errors since last reset.
 *   Bytes 16-19: Number of frame overrun errors since last reset.
 *   Bytes 20-23: Number of frame length errors since last reset.
 *   Bytes 24-27: Number of frames received since last reset.
 *   Bytes 28-31: Number of frames transmitted since last reset.
 *
 * 1-byte counters have a value between 0 and 255. 4-byte counters are
 * transmitted LSB first and have a value between 0 and 4,294,967,295.
 *
 * The RS-232 port control signals are used as follows:
 *	DTR - RS-232 driver/receiver enable
 *	DSR - RS-232 signal level invalid
 *	DCD - RS-232 signal level invalid
 *	RTS - RS-232 output
 *	CTS - RS-232 input
 *
 * PIO port E is used to control the RS-485 transceiver:
 *	Bit 3 - Receiver disable
 *	Bit 4 - Driver enable
 *
 * Event flags are used for communication between the ISR and driver
 * and between the driver and user programs:
 *
 *	Bit 0 - Data received on port 0
 *	Bit 1 - Data received on port 1
 *	Bit 2 - Data received on port 2
 *	Bit 3 - ISR completed transmission of one frame
 *
 * The following buffers are used for interprocess communication:
 *
 *	TXBUFFER - Set by driver to point to data buffer to be transmitted.
 *	           Set to 0 by ISR when transmission complete
 *	RXBUFFER - Set by driver to point to data buffer for ISR.
 *	           Set to 0 by ISR when frame received.
 *	TXPOOL - Driver places transmitted and acknowledged data buffers
 *	         into this queue for reuse by user programs.
 *	RXPOOL - Driver gets buffers for incoming fames from this queue.
 *	         Driver and user program place discarded buffers into pool.
 *	TXQ[] - Queues of data buffers to be transmitted on each port.
 *	        Filled by user programs and emptied by driver.
 *	RXQ[] - Queues of data buffers with data received by driver.
 *	        Read by user program for each port.
 *	RXF[] - Semaphore for each port showing number of frames to receive.
 *	        Incremented by user programs and decremented by driver.
 *
 *                                             +-----------+
 *                                    +------->|  RXF[0-2] |<---------+
 *                                    |      - +-----------+ +        |
 *                                    |                               |
 *                                    |        +-----------+          |
 *                                    |  +-----|  RXPOOL   |<------+  |
 *                                    |  |     +-----------+       |  |
 *                                    |  V                       +---------+
 * +---------+   +-----------+   +---------+   +-----------+   +---------+ |
 * |         |-->| RXBUFFER  |-->|         |-->|  RXQ[0-2] |-->|         | |
 * |         |   +-----------+   |         |   +-----------+   |  USER   | |
 * |   ISR   |                   | DRIVER  |                   | PROGRAM | |
 * |         |   +-----------+   |         |   +-----------+   |         | |
 * |         |<--| TXBUFFER  |<--|         |<--|  TXQ[0-2] |<--|         |-+
 * +---------+   +-----------+   +---------+   +-----------+   +---------+
 *                                    |                             A
 *                                    |        +-----------+        |
 *                                    +------->|  TXPOOL   |--------+
 *                                             +-----------+
 *
 * Note that user programs must allow one byte at the beginning oe each buffer
 * to allow the protocol to insert the header and one byte at the end of each
 * buffer for the parity byte.
 */

#include    "NML675001.h"
#include    "DCP1-RTOS.h"
#include    "queues.h"

/* UART driver PCB */
int uart_pcb[32];

/* pool */
int poolram[30*19];	/* buffer area */
queue pool;		/* queue control structures */

/* fixed buffer for control frames - matches queues.h definition */
struct {
	int link;	/* unused */
	int size;	/* frame length in bytes */
	char data[68];	/* frame contents */
} control;

/* transmit and receive queues and state info for each port */
queue txq[3], rxq[3];	/* input and output queues */
int txs[3], rxs[3];	/* transmitter and receiver state */
int rxf[3];		/* receiver flow control */

/* global variables */
char myaddress;		/* local address in upper 4 bits - lower 4 are 0 */

/* statistics counters */
int	rxinterrupts, txinterrupts, 
	byteparity, frameparity, byteoverrun, frameoverrun, lengtherror, 
	bytesreceived, bytestransmitted, framesreceived, framestransmitted;

/* ISR state */
buffer *txbuffer, *rxbuffer;	/* pointers to filled transmit and receive buffers */
int txindex, rxindex;		/* next character to transmit */
char rxesc;			/* set to 20H when escape (125) detected or generated */

/*
 * Transmit frame
 *
 * Adds parity to frame, flushes FIFO, enables interrupt, initializes ISR and sends flag.
 * Then waits for event signalling last bytes put in FIFO and returns. If half-duplex
 * (RS-485) link, waits for transmitter shift register to empty before returning.
 */
void txframe (buffer* g)
{
	int i;
	
	if (g == 0) return;	/* validate parameter */
	/* calculate parity */
	i = g->size - 1;
	g->data[i] = parity(g->data, i);
	/* set ISR state */
	txindex = 0;	/* reset index */
	txbuffer = g;	/* start data transmission */
	/* set up UART transmitter */
	if (UARTMSR & 0x80)	/* check for DCD on (invalid RS-232 levels */
	{
		/* yes, use RS-485 */
		GPPOE |= 0x0018;	/* RXEN- = 1, TXEN+ = 1 */
		UARTFCR = 0x85;		/* clear transmit FIFO */
		UARTFCR = 0x81;		/* TX & RX FIFO on & trigger = 8 */
/*		UARTTHR = 126;		/* send opening flag */
		UARTIER = 0x02;		/* tx int on and rx int off */
	} else {
		/* no, use RS-232 levels */
		UARTMCR = 0x03;		/* DTR & RTS on (low) */
/*		UARTTHR = 126;		/* send opening flag */
		UARTIER |= 0x02;	/* set bit 2 - tx int on */
	}; /* end if */
	/* wait for end of transmission event flag */
	#asm
	swi	0
	#endasm
	/* if DCD on (invalid RS-232 levels) wait for transmitter to empty */
	if ((UARTMSR & 0x80) && ((UARTLSR & UARTLSR_TEMT) == 0))
	{
		/* wait for transmitter shift register to empty */
		/* this wastes CPU cycles but the TXE flag has no interrupt */
		do ; while((UARTLSR & UARTLSR_TEMT) == 0);
	}; /* end if */	
}; /* end txframe() */

/*
 * Receive frame
 *
 * Flushes FIFO, enables interrupt and initializes ISR. Then waits for event flag
 * signalling frame received and checks parity and address. Good frames addressed
 * to this station are returned. 
 */
void rxframe (buffer* f)
{
	int p;
	
	if (f == 0) return;	/* validate parameter */
	/* receive frames until one found for this address without error */
	do {
		/* initialize ISR */
		rxindex = 0;	/* 0 causes wait for first non-flag character */
		rxesc = 0;	/* cancel escape detect */
		rxbuffer = f;	/* start accumulating frame */
		/* enable interrupt */
		if (UARTMSR & 0x80)	/* check for DCD on (invalid RS-232 levels) */
		{
			/* yes, use RS-485 */
			GPPOE &= 0xffe7;	/* RXEN- = 0, TXEN+ = 0 */
			UARTFCR = 0x83;		/* clear receive FIFO */
			UARTFCR = 0x81;		/* TX & RX FIFO on & trigger = 8 */
			UARTIER = 0x01;		/* rx int on and tx int off */
		} else {
			/* no, use RS-232 levels */
			UARTMCR = 0x03;		/* DTR and RTS on (low) */
			UARTIER |= 0x01;	/* set bit 1 - rx int on */
		}; /* end if */
		/* wait for event */
		#asm
		swi	0
		#endasm
		/* check parity */
		p = parity(f->data, f->size);
		if (p == 0)
			framesreceived++;
		else
			frameparity++;
	} while	((p != 0) || ((f->data[0] & 0xF0) != myaddress));
}; /* end rxframe() */

/*
 * Configure UART for specified baud rate
 *
 * Address = 4-bit local address
 * Buad rate = CCLK/divisor/16
 */
void uart_init(int address, int divisor)
{
	int i;
	/* set address */
	myaddress = (char)(address << 4);
	/* initialize UART hardware */
	UARTLCR = 0x9B;	/* 8-bits, even parity, 1 stop bit */
	UARTDLL = (unsigned char) divisor;	/* set baud rate */
	UARTDLM = (unsigned char) (divisor >> 8);
	UARTLCR = 0x1B;	/* divisor register access off */
	UARTIER = 0x00;	/* all interrupts off */
	UARTMCR = 0;	/* RS-232 off */
	UARTFCR = 0x87;	/* clear tx and rx FIFOs */
	UARTFCR = 0x81;	/* TX & RX FIFO on & trigger = 8 */
	GPPME = 0x0000;	/* bits 3 and 4 are output */
	GPIEE = 0x0000;	/* no interrupts */
	GPPOE = (GPPOE & 0xffe7) | 0x0008;	/* RXEN- = 1, TXEN+ = 0 */
	/* zero error counters */
	rxinterrupts = 0;
	txinterrupts = 0;
	byteparity = 0;
	frameparity = 0;
	byteoverrun = 0;
	frameoverrun = 0;
	lengtherror = 0;
	bytesreceived = 0;
	bytestransmitted = 0;
	framesreceived = 0;
	framestransmitted = 0;
	/* initialize transmit and receive queues */
	for (i = 2; i >= 0; --i)
	{
		qnew(&txq[i]);
		qnew(&rxq[i]);
	}; /* end for */
	/* create buffer pool */
	qfill(&pool,poolram,30,19);
	/* initialize sequence numbers and flow control */
	for (i = 2; i >= 0; --i)
	{
		txs[i] = 0;
		rxs[i] = 0;
		rxf[i] = 0;
	}; /* end for */
	/* reset ISR state */
	rxbuffer = 0;	/* no buffer */
	txbuffer = 0;	/* no buffer */
	rxindex = 32767;	/* end of frame */
	txindex = 32767;	/* end of frame */
	/* connect UART ISR to IRN 9 with IRL=6 */
	#asm
	load	r0, #9
	load	r1, #6
	load	r2, #_uart_isr
	swi	4
	#endasm
	/* register protocol driver with priority 12 and event flag 3 */
	#asm
	load	r0, #12
	load	r1, #0x08
	load	r2, #_uart_pcb+124
	load	r3, #_uart_main
	swi	1
	#endasm
	/* start protocol driver by setting event flag 3 */
	#asm
	load	r0, #3
	swi	6
	#endasm
}; /* end uart_init() */

/*
 * UART Interrupt Service Routine
 *
 * Fills rxbuffer with any bytes received between flag characters while removing
 * escape sequences. Transmits txbuffer inserting escape sequences and appends flag
 * character.
 */
void uart_isr(void)
{
	char c,s,esc;	/* character, UART line status and escape sequence flag */
	buffer *f;	/* pointer to buffer */
	char *base;	/* base address of buffer data field */
	int i;		/* current index into receive or transmit buffer */
	
	/* get interrupt source */
	switch (UARTIIR & 0x0E)
	{
		case 6:		/* receiver line status */
		s = UARTLSR;	/* reset interrupt */
		break;
		
		case 4: 	/* receiver FIFO trigger level reached */
		case 12:	/* receiver timeout */
		rxinterrupts++;
		/* get receive buffer */
		f = rxbuffer;
		if (f == 0)
		{
			if (UARTRBR != 126)	/* drop character if no buffer */
				frameoverrun++;	/* error if not flag */		
		} else {
			base = f->data;	/* retreive data buffer start address */
			esc = rxesc;	/* get escape flag */
			i = rxindex;	/* get current position in buffer */
			for (s = UARTLSR & 0x1F; (s != 0) && (f != 0); s = UARTLSR & 0x1F)
			{
				c = UARTRBR;	/* get character */
				if (s & 0x1E)	/* check for errors */
				{
					base[i++] = 127;	/* substitute abort for erroneous characters */
					if (s == 0x02)		/* check for overrun error */
						byteoverrun++;	/* report overrun error */
					else
						byteparity++;	/* framing error & break reported as parity error */
				} else {
					bytesreceived++;
					switch (c)	/* process good characters */
					{
						case 125:	/* escape character */
						esc = 0x20;	/* inver bit 5 of next received character */
						break;
						
						case 126:	/* flag character */
						if (i != 0)	/* if closing flag */
						{
							if (i <= 66)	/* check frame length */
							{
								f->size = i;	/* save length */
								rxbuffer = 0;	/* send buffer to driver */
								f = 0;		/* terminate loop */
							} else {
								lengtherror++;	/* report error */
							}; /* end if */
							i = 0;		/* start new frame */
						}; /* end if */
						esc = 0;	/* flag ends escape sequence */
						break;
						
						default:	/* normal character */
						if (i <= 66)	/* if maximum frame length not exceeded */
							base[i++] = c ^ esc;	/* save modified character */
						esc = 0;	/* reset escape flag */
						break;
					}; /* end switch */
				}; /* end if */
			}; /* end for */
			rxindex = i;	/* save position in buffer */
			rxesc = esc;	/* save escape sequence flag */
			if (f == 0)	/* if end of frame */
			{
				UARTIER &= 0xFE; /* stop interrupts */
				/* signal driver */
				#asm
				load	r0, #3
				swi	6
				#endasm
			}; /* end if */						
		}; /* end if */
		break;
		
		case 2:		/* transmit queue empty */
		txinterrupts++;
		/* get transmit buffer */
		f = txbuffer;	/* address of buffer */
		if (f == 0)
		{
			UARTTHR = 126;	/* send flag if no buffer */		
		} else {		
			base = f->data;	/* address of data field */
			s = f->size;	/* frame length */
			i = txindex;	/* current position */
			do {
				bytestransmitted++;
				if (i < s)	/* check location in field */
				{
					c = base[i++];	/* get character to transmit */
					if ((c >= 125) && (c <= 127)) /* control character? */
					{									
						UARTTHR = 125;	/* send escape character */
						UARTTHR = c ^ 0x20;
					} else {
						UARTTHR = c;	/* send data */
					}; /* end if */
				} else {
					UARTTHR = 126;	/* opening and closing flags */
					UARTTHR = 126;	/* opening and closing flags */
					UARTTHR = 126;	/* opening and closing flags */
					UARTTHR = 126;	/* opening and closing flags */
					i++;
				}; /* end if */
			} while ((UARTLSR & 0x20) && (i <= s));	/* done if no space in FIFO or buffer transmitted */
			txindex = i;	/* save current position */
			if (i > s)	/* check for transmission done */
			{
				framestransmitted++;	/* count frame */
				txindex = 0;		/* reset current position */
				txbuffer = 0;		/* indicate ISR ready for another buffer */
				UARTIER &= 0xFD;	/* reset bit 2 - TX IE OFF */
				/* signal driver */
				#asm
				load	r0, #3
				swi	6
				#endasm
			}; /* end if */
		}; /* end if */
		break;
		
		case 0:		/* modem status */
		s = UARTMSR;	/* reset interrupt */
		break;
	}; /* end switch on UARTIIR */
}; /* end uart_isr() */

/*
 * Driver
 *
 * Receives frames and distributes to user programs.
 * Transmits accumulated frames when polled.
 */
void uart_main(void)
{
	buffer *f,*g;	/* receive and transmit buffer pointers */
	int port;	/* port field value */
	int sn;		/* sequence field value */
	int pa;		/* poll/acknowledge field */
	int i; 		/* loop counter */
	
	for (;;)	/* loop forever */
	{
		/* get buffer */
		f = qnext(&pool);
		if (f != 0)
		{		
			/* receive frame and decode */
			rxframe(f);
			port = f->data[0] & 0x03;
			sn = f->data[0] & 0x04;
			pa = f->data[0] & 0x08;
			/* check port number */
			if (port == 3)	/* process special frames */
			{
				/* by default, no data in response */
				control.size = 2;
				control.data[0] = f->data[0];
				switch (f->data[0] & 0xC0)	/* check PA and SQ bits */
				{
					case 0x00:	/* reset all ports */
					/* reset statistics counters */
					rxinterrupts = 0;
					txinterrupts = 0;
					byteparity = 0;
					frameparity = 0;
					byteoverrun = 0;
					frameoverrun = 0;
					lengtherror = 0;
					bytesreceived = 0;
					bytestransmitted = 0;
					framesreceived = 0;
					framestransmitted = 0;
					/* reset state variables, flush queues and signal users */
					for (i = 2; i >= 0; --i)
					{
						txs[i] = 0;	/* reset link state */
						rxs[i] = 0;
						semaphore(&rxf[i],0);	/* reset low control */
						qmlast(&pool, &txq[i]);	/* flush transmit and */
						qmlast(&pool, &rxq[i]);	/* receive queues */
						ossetflg(i);
					}; /* end for */
					break;
					
					case 0x40:	/* get statistics */
					control.size = 40;
					control.data[1] = 0;
					control.data[2] = (char)qcount(&pool);
					control.data[3] = (char)qcount(&pool);
					putint(&control.data[4], byteparity);
					putint(&control.data[8], frameparity);
					putint(&control.data[12], byteoverrun);
					putint(&control.data[16], frameoverrun);
					putint(&control.data[20], lengtherror);
					putint(&control.data[24], bytesreceived);
					putint(&control.data[28], bytestransmitted);
					putint(&control.data[32], framesreceived);
					putint(&control.data[36], framestransmitted);
					break;
					
					case 0x80:
					break;
					
					case 0xC0:
					break;
				}; /* end switch */
				qlast(&pool,f);	/* discard received frame */
				txframe((buffer*)&control);	/* post frame for transmission */
			}
			else	/* or process normal frames */
			{
				if (pa == 0)	/* check for data frame */
				{
					/* check sequence number */
					if (sn == rxs[port])
					{
						/* check flow control */
						if (semaphore(&rxf[port],-1) > 0)
						{
							/* and deliver if space */
							qlast(&rxq[port],f);
							/* increment expected sn */
							rxs[port] ^= 0x04;
							/* and notify user program */
							ossetflg(port);
							/* acknowledge data frame */
							control.data[0] = 0x08 | sn | port | myaddress;
						} else {
							/* discard data frame */
							qlast(&pool,f);
							/* don't acknowledge data frame */
							control.data[0] = 0x08 | (sn ^ 0x04) | port | myaddress;
						}; /* end if */
					} else {
						/* discard if duplicate */
						qlast(&pool,f);
						/* and acknowledge */
						control.data[0] = 0x08 | sn | port | myaddress;
					}; /* end if */
					/* send acknowledgement frame */
					control.size = 2;
					txframe((buffer*)&control);
				}
				else	/* or polling frame */
				{
					if (sn != txs[port])	/* check sequence number */
					{
						/* discard old frame if new one requested */
						qlast(&pool,qnext(&txq[port]));
						/* set state to new value */
						txs[port] = sn;
					}; /* end if */
					qlast(&pool,f);	/* discard received poll */
					g = qcheck(&txq[port]);	/* check data availability */
					if (g != 0)
					{	/* add header and send data frame */
						g->data[0] = sn | port | myaddress;
						txframe(g);
					} else {
						/* send no data available frame */
						control.size = 2;
						control.data[0] = (sn ^ 0x04) | port | myaddress;
						txframe((buffer*)&control);
					}; /* end if */
				}; /* end if */
			}; /* end if */
		}; /* end if */		
	}; /* end for */
}; /* end uart_main */

/*
 * The following functions provide byte-wide I/O
 */
int urdp[3], uwrp[3];		/* read and write pointers */
buffer *urdb[3], *uwrb[3];	/* read and write buffers */

/*
 * Reset character counters and buffer pointers.
 * Discard any buffers in use.
 */
void ureset(int port)
{
	if ((port > 2) || (port < 0)) return;
	urdb[port] = 0;
	urdp[port] = 0;
	uwrb[port] = 0;		
	uwrp[port] = 0;		
}; /* end ureset() */

/*
 * Read character from selected port.
 * Returns character value or zero if no character ready.
 */
int ugetc(int port, char *c)
{
	int i;
	buffer *f;
	/* validate parameters */
	if ((port > 2) || (port < 0)) return 0;
	/* check for buffer */
	f = urdb[port];
	i = urdp[port];
	if (f == 0)
	{
		/* and get one if not present */
		i = semaphore(&rxf[port],1);
		f = qnext(&rxq[port]);
		if (f == 0) return 0;
		i = 1;	/* first byte in user data field */
	}; /* end if */
	/* get next character */		
	*c = f->data[i++];
	/* discard buffer if last character */
	if (i >= (f->size - 1))
	{
		qlast(&pool,f);
		f = 0;
	}; /* end if */		
	/* save pointer and counter and return */
	urdb[port] = f;
	urdp[port] = i;
	return 1;
}; /* end ugetc() */

/*
 * Write character to selected port.
 * Returns -1 if success.
 */
int uputc(int port, char c)
{
	int i;
	buffer *g;
	/* validate parameters */
	if ((port > 2) || (port < 0)) return 0;
	/* check for buffer */
	g = uwrb[port];
	i = uwrp[port];
	if (g == 0)
	{
		/* and get one if not present */
		g = qnext(&pool);
		if (g == 0) return 0;
		i = 1;	/* first byte in user data field */
	}; /* end if */
	/* put character in buffer */		
	g->data[i++] = c;
	/* transmit buffer if last character */
	if (i >= 65)
	{
		g->size = 66;
		qlast(&txq[port],g);
		g = 0;
	}; /* end if */		
	/* save pointer and counter and return */
	uwrb[port] = g;
	uwrp[port] = i;
	return 1;
}; /* end uputc() */

/*
 * Disard remaining characters in receiver buffer.
 * Send remaining characters in transmit buffer.
 */
void uflush(int port)
{
	int i;
	buffer *f,*g;
	/* validate parameters */
	if ((port > 2) || (port < 0)) return;
	/* check for input buffer and discard */
	f = urdb[port];
	if (f == 0)
		f = qnext(&rxq[port]);
	if (f != 0)
	{
		do {	/* empty receive queue */
			qlast(&pool,f);
			f = qnext(&rxq[port]);
		} while (f != 0);
	}; /* end if */	
	urdb[port] = 0;
	urdp[port] = 0;
	/* check for output buffer */
	g = uwrb[port];
	i = uwrp[port];
	if (g == 0) return;	/* doneif none */
	/* else transmit buffer */
	g->size = i+1;
	qlast(&txq[port],g);
	uwrb[port] = 0;
	uwrp[port] = 0;
}; /* end uflush() */

